import subprocess
import os


def get_syspath(devpath: str):
    for line in subprocess.run(
        ["udevadm", "info", devpath], capture_output=True
    ).stdout.splitlines():
        if line.startswith(b"P: "):
            return line[3:].decode()
    return None


def get_gamepad_name(devpath: str):
    syspath = get_syspath(devpath)
    if not syspath:
        return None

    parts = syspath.split("/")
    if len(parts) < 3:
        return None
    input_dev = parts[-2]
    if not input_dev.startswith("input") or input_dev == "input":
        return None
    return input_dev


def get_parent_sysfs(devpath: str):
    syspath = get_syspath(devpath)
    if not syspath:
        return None

    return syspath[: syspath.rindex("/")]


def reload_children(parent: str):
    stat = subprocess.run(
        ["udevadm", "control", "--reload-rules"],
        capture_output=True,
    )
    if stat.returncode:
        return False
    for action in ["remove", "add"]:
        stat = subprocess.run(
            ["udevadm", "trigger", "--action", action, "-b", parent],
            capture_output=True,
        )
        if stat.returncode:
            return False
    return True


def hide_gamepad(devpath: str, vid: int, pid: int):
    input_dev = get_gamepad_name(devpath)
    parent = get_parent_sysfs(devpath)
    if not input_dev or not parent:
        return False

    rule = f"""\
# Hides device gamepad devices stemming from {input_dev}
# Managed by HHD, this file will be autoremoved during configuration changes.
SUBSYSTEMS=="input", KERNELS=="{input_dev}", ATTRS{{id/vendor}}=="{vid:04x}", ATTRS{{id/product}}=="{pid:04x}", GOTO="hhd_valid"
GOTO="hhd_end"
LABEL="hhd_valid"
KERNEL=="js[0-9]*|event[0-9]*", SUBSYSTEM=="input", MODE="000", GROUP="root", TAG="", RUN+="/bin/chmod 000 /dev/input/%k"
LABEL="hhd_end"
"""  # , RUN+="/bin/chmod 000 /sys/%p"
    try:
        os.makedirs("/run/udev/rules.d/", exist_ok=True)
        with open(f"/run/udev/rules.d/95-hhd-devhide-{input_dev}.rules", "w") as f:
            f.write(rule)
        return reload_children(parent)
    except Exception:
        return False


def unhide_gamepad(devpath: str):
    input_dev = get_gamepad_name(devpath)
    parent = get_parent_sysfs(devpath)
    if not input_dev or not parent:
        return False

    try:
        os.remove(f"/run/udev/rules.d/95-hhd-devhide-{input_dev}.rules")
        return reload_children(parent)
    except Exception:
        return False


def unhide_all():
    try:
        for rule in os.listdir("/run/udev/rules.d/"):
            if rule.startswith("95-hhd-devhide"):
                os.remove(os.path.join("/run/udev/rules.d/", rule))
    except Exception:
        pass



from typing import Any, Mapping, Sequence

from .utils import load_relative_yaml
from ..controller.base import Consumer, Producer
from ..controller.virtual.dualsense import Dualsense, TouchpadCorrectionType
from ..controller.virtual.uinput import (
    HHD_PID_MOTION,
    HHD_PID_TOUCHPAD,
    MOTION_AXIS_MAP,
    MOTION_CAPABILITIES,
    TOUCHPAD_AXIS_MAP,
    TOUCHPAD_BUTTON_MAP,
    TOUCHPAD_CAPABILITIES,
    UInputDevice,
)


def get_outputs(
    conf, touch_conf, motion: bool = False
) -> tuple[Sequence[Producer], Sequence[Consumer], Mapping[str, Any]]:
    producers = []
    consumers = []

    controller = conf["mode"].to(str)
    if touch_conf is not None:
        touchpad = touch_conf["mode"].to(str)
        correction = touch_conf["controller.correction"].to(TouchpadCorrectionType)
    else:
        touchpad = "controller"
        correction = "stretch"

    uses_touch = False
    uses_leds = False
    match controller:
        case "dualsense_edge":
            uses_touch = touchpad == "controller"
            uses_leds = conf.get("dualsense_edge.led_support", False)
            d = Dualsense(
                touchpad_method=correction,
                edge_mode=True,
                use_bluetooth=conf["dualsense_edge.bluetooth_mode"].to(bool),
                enable_touchpad=uses_touch,
                enable_rgb=uses_leds,
                fake_timestamps=not motion,
                sync_gyro=conf["dualsense_edge.sync_gyro"].to(bool) and motion,
                paddles_to_clicks=False,
            )
            producers.append(d)
            consumers.append(d)
        case "dualsense":
            uses_touch = touchpad == "controller"
            uses_leds = conf.get("dualsense.led_support", False)
            d = Dualsense(
                touchpad_method=correction,
                edge_mode=False,
                use_bluetooth=conf["dualsense.bluetooth_mode"].to(bool),
                enable_touchpad=uses_touch,
                enable_rgb=uses_leds,
                fake_timestamps=not motion,
                sync_gyro=conf["dualsense.sync_gyro"].to(bool) and motion,
                paddles_to_clicks=conf["dualsense.paddles_to_clicks"].to(bool),
            )
            producers.append(d)
            consumers.append(d)
        case "uinput":
            d = UInputDevice(phys="phys-hhd-main")
            producers.append(d)
            consumers.append(d)
            if motion:
                d = UInputDevice(
                    name="Handheld Daemon Controller Motion Sensors",
                    phys="phys-hhd-main",
                    capabilities=MOTION_CAPABILITIES,
                    pid=HHD_PID_MOTION,
                    btn_map={},
                    axis_map=MOTION_AXIS_MAP,
                    output_imu_timestamps=True,
                )
                producers.append(d)
                consumers.append(d)
        case _:
            raise RuntimeError(f"Invalid controller type: '{controller}'.")

    if touchpad == "emulation":
        d = UInputDevice(
            name="Handheld Daemon Touchpad",
            phys="phys-hhd-main",
            capabilities=TOUCHPAD_CAPABILITIES,
            pid=HHD_PID_TOUCHPAD,
            btn_map=TOUCHPAD_BUTTON_MAP,
            axis_map=TOUCHPAD_AXIS_MAP,
            output_timestamps=True,
        )
        producers.append(d)
        consumers.append(d)
        uses_touch = True

    return (
        producers,
        consumers,
        {"uses_touch": uses_touch, "uses_leds": uses_leds, "is_dual": False},
    )


def get_outputs_config(
    can_disable: bool = False,
    has_leds: bool = True,
    start_disabled: bool = False,
    default_device: str | None = None,
):
    s = load_relative_yaml("outputs.yml")
    if not can_disable:
        del s["modes"]["disabled"]
    if not has_leds:
        del s["modes"]["dualsense"]["children"]["led_support"]
        del s["modes"]["dualsense_edge"]["children"]["led_support"]

    if default_device:
        s["default"] = default_device
    if start_disabled:
        s["default"] = "disabled"
    return s



import argparse
import logging
import re
import select
import sys
import time
from threading import Event as TEvent
from typing import Literal, Sequence

from hhd.controller import Button, Consumer, Event, Producer, Axis
from hhd.controller.base import Multiplexer, TouchpadAction
from hhd.controller.lib.hid import enumerate_unique
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev
from hhd.controller.physical.hidraw import GenericGamepadHidraw
from hhd.controller.physical.imu import AccelImu, GyroImu
from hhd.controller.virtual.uinput import HHD_PID_VENDOR, UInputDevice
from hhd.plugins import Config, Context, Emitter, get_outputs

from .const import (
    LGO_RAW_INTERFACE_AXIS_MAP,
    LGO_RAW_INTERFACE_BTN_ESSENTIALS,
    LGO_RAW_INTERFACE_BTN_MAP,
    LGO_RAW_INTERFACE_CONFIG_MAP,
    LGO_TOUCHPAD_AXIS_MAP,
    LGO_TOUCHPAD_BUTTON_MAP,
)
from .gyro_fix import GyroFixer
from .hid import RgbCallback

ERROR_DELAY = 1
SELECT_TIMEOUT = 1

logger = logging.getLogger(__name__)

LEN_VID = 0x17EF
LEN_PIDS = {
    0x6182: "xinput",
    0x6183: "dinput",
    0x6184: "dual_dinput",
    0x6185: "fps",
}


def plugin_run(
    conf: Config, emit: Emitter, context: Context, should_exit: TEvent, updated: TEvent
):
    # Remove leftover udev rules
    # unhide_all()
    if (gyro_fix := conf.get("gyro_fix", False)) and conf["gyro"].to(bool):
        gyro_fixer = GyroFixer(int(gyro_fix) if int(gyro_fix) > 10 else 100)
    else:
        gyro_fixer = None

    while not should_exit.is_set():
        try:
            controller_mode = None
            pid = None
            while not controller_mode:
                devs = enumerate_unique(LEN_VID)
                if not devs:
                    logger.error(
                        f"Legion go controllers not found, waiting {ERROR_DELAY}s."
                    )
                    time.sleep(ERROR_DELAY)
                    continue

                for d in devs:
                    if d["product_id"] in LEN_PIDS:
                        pid = d["product_id"]
                        controller_mode = LEN_PIDS[pid]
                        break
                else:
                    logger.error(
                        f"Legion go controllers not found, waiting {ERROR_DELAY}s."
                    )
                    time.sleep(ERROR_DELAY)
                    continue

            conf_copy = conf.copy()
            updated.clear()
            if (
                controller_mode == "xinput"
                and conf["xinput.mode"].to(str) != "disabled"
            ):
                logger.info("Launching emulated controller.")
                if gyro_fixer:
                    gyro_fixer.open()
                controller_loop_xinput(conf_copy, should_exit, updated)
            else:
                if controller_mode != "xinput":
                    logger.info(
                        f"Controllers in non-supported (yet) mode: {controller_mode}."
                    )
                else:
                    logger.info(
                        f"Controllers in xinput mode but emulation is disabled."
                    )
                controller_loop_rest(
                    controller_mode, pid if pid else 2, conf_copy, should_exit, updated
                )
        except Exception as e:
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {ERROR_DELAY}s."
            )
            # Raise exception
            if conf.get("debug", False):
                raise e
            time.sleep(ERROR_DELAY)
        finally:
            if gyro_fixer:
                gyro_fixer.close()
            # Remove leftover udev rules
            # unhide_all()


def controller_loop_rest(
    mode: str, pid: int, conf: Config, should_exit: TEvent, updated: TEvent
):
    debug = conf.get("debug", False)
    shortcuts_enabled = conf["shortcuts"].to(bool)
    # FIXME: Sleep when shortcuts are disabled instead of polling raw interface
    if shortcuts_enabled:
        logger.info(f"Launching a shortcuts device.")
    else:
        logger.info(f"Shortcuts disabled. Waiting for controllers to change modes.")

    d_raw = SelectivePassthrough(
        GenericGamepadHidraw(
            vid=[LEN_VID],
            pid=list(LEN_PIDS),
            usage_page=[0xFFA0],
            usage=[0x0001],
            report_size=64,
            axis_map=LGO_RAW_INTERFACE_AXIS_MAP,
            btn_map=LGO_RAW_INTERFACE_BTN_MAP,
            required=True,
        )
    )

    multiplexer = Multiplexer(
        dpad="analog_to_discrete",
        trigger="analog_to_discrete",
        share_to_qam=conf["share_to_qam"].to(bool),
        nintendo_mode=conf["nintendo_mode"].to(bool),
    )
    d_uinput = UInputDevice(
        name=f"HHD Shortcuts (Legion Mode: {mode})",
        pid=HHD_PID_VENDOR | 0x0200 | (pid & 0xF),
        phys=f"phys-hhd-shortcuts-legion-{mode}",
    )

    d_shortcuts = GenericGamepadEvdev(
        vid=[LEN_VID],
        pid=list(LEN_PIDS),
        name=[re.compile(r"Legion-Controller \d-.. Keyboard")],
        capabilities={EC("EV_KEY"): [EC("KEY_1")]},
        required=True,
    )

    try:
        fds = []
        fds.extend(d_raw.open())
        if shortcuts_enabled:
            fds.extend(d_shortcuts.open())
            fds.extend(d_uinput.open())

        while not should_exit.is_set() and not updated.is_set():
            select.select(fds, [], [], SELECT_TIMEOUT)
            evs = multiplexer.process(d_raw.produce(fds))

            if shortcuts_enabled:
                d_shortcuts.produce(fds)
                d_uinput.produce(fds)
                if debug and evs:
                    logger.info(evs)
                d_uinput.consume(evs)
    finally:
        d_uinput.close(True)
        d_shortcuts.close(True)
        d_raw.close(True)


def controller_loop_xinput(conf: Config, should_exit: TEvent, updated: TEvent):
    debug = conf.get("debug", False)

    # Output
    motion = conf["accel"].to(bool) or conf["gyro"].to(bool)
    d_producers, d_outs, d_params = get_outputs(
        conf["xinput"], conf["touchpad"], motion
    )

    # Imu
    d_accel = AccelImu()
    # Legion go has a bit lower sensitivity than it should
    GYRO_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
        "anglvel_x": ("gyro_z", "anglvel", -conf["gyro_scaling"].to(int), None),
        "anglvel_y": ("gyro_x", "anglvel", conf["gyro_scaling"].to(int), None),
        "anglvel_z": ("gyro_y", "anglvel", conf["gyro_scaling"].to(int), None),
        "timestamp": ("gyro_ts", None, 1, None),
    }
    d_gyro = GyroImu(map=GYRO_MAPPINGS)

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[0x17EF],
        pid=[0x6182],
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
    )
    d_touch = GenericGamepadEvdev(
        vid=[0x17EF],
        pid=[0x6182],
        name=[re.compile(".+Touchpad")],  # "  Legion Controller for Windows  Touchpad"
        capabilities={EC("EV_KEY"): [EC("BTN_MOUSE")]},
        btn_map=LGO_TOUCHPAD_BUTTON_MAP,
        axis_map=LGO_TOUCHPAD_AXIS_MAP,
        aspect_ratio=1,
        required=True,
    )
    d_raw = SelectivePassthrough(
        GenericGamepadHidraw(
            vid=[LEN_VID],
            pid=list(LEN_PIDS),
            usage_page=[0xFFA0],
            usage=[0x0001],
            report_size=64,
            axis_map=LGO_RAW_INTERFACE_AXIS_MAP,
            btn_map=LGO_RAW_INTERFACE_BTN_MAP,
            config_map=LGO_RAW_INTERFACE_CONFIG_MAP,
            callback=RgbCallback(),
            required=True,
        )
    )
    # Mute keyboard shortcuts, mute
    d_shortcuts = GenericGamepadEvdev(
        vid=[LEN_VID],
        pid=list(LEN_PIDS),
        name=[re.compile(".+Keyboard")],  # "  Legion Controller for Windows  Keyboard"
        # capabilities={EC("EV_KEY"): [EC("KEY_1")]},
        # report_size=64,
        required=True,
    )

    match conf["swap_legion"].to(str):
        case "disabled":
            swap_guide = None
        case "l_is_start":
            swap_guide = "guide_is_start"
        case "l_is_select":
            swap_guide = "guide_is_select"
        case val:
            assert False, f"Invalid value for `swap_legion`: {val}"

    touch_actions = (
        conf["touchpad.controller"]
        if conf["touchpad.mode"].to(TouchpadAction) == "controller"
        else conf["touchpad.emulation"]
    )

    multiplexer = Multiplexer(
        swap_guide=swap_guide,
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        led="main_to_sides",
        status="both_to_main",
        share_to_qam=conf["share_to_qam"].to(bool),
        touchpad_short=touch_actions["short"].to(TouchpadAction),
        touchpad_right=touch_actions["hold"].to(TouchpadAction),
        select_reboots=conf["select_reboots"].to(bool),
        r3_to_share=conf["m2_to_mute"].to(bool),
        nintendo_mode=conf["nintendo_mode"].to(bool),
    )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        prepare(d_xinput)
        if conf.get("accel", False):
            prepare(d_accel)
        if conf.get("gyro", False):
            prepare(d_gyro)
        prepare(d_shortcuts)
        if d_params["uses_touch"]:
            prepare(d_touch)
        prepare(d_raw)
        for d in d_producers:
            prepare(d)

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                d_xinput.consume(evs)
                d_raw.consume(evs)

            for d in d_outs:
                d.consume(evs)

            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        for d in reversed(devs):
            d.close(True)


class SelectivePassthrough(Producer, Consumer):
    def __init__(
        self,
        parent,
        forward_buttons: Sequence[Button] = ("share", "mode"),
        passthrough: Sequence[Button] = list(
            next(iter(LGO_RAW_INTERFACE_BTN_ESSENTIALS.values()))
        ),
    ):
        self.parent = parent
        self.state = False

        self.forward_buttons = forward_buttons
        self.passthrough = passthrough

        self.to_disable_btn = set()
        self.to_disable_axis = set()

    def open(self) -> Sequence[int]:
        return self.parent.open()

    def close(self, exit: bool) -> bool:
        return self.parent.close(exit)

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        evs: Sequence[Event] = self.parent.produce(fds)

        out = []
        prev_state = self.state
        for ev in evs:
            if ev["type"] == "button" and ev["code"] in self.forward_buttons:
                self.state = ev.get("value", False)

            if ev["type"] == "configuration":
                out.append(ev)
            elif ev["type"] == "button" and ev["code"] in self.passthrough:
                out.append(ev)
            elif ev["type"] == "button" and self.state:
                self.to_disable_btn.add(ev["code"])
            elif ev["type"] == "axis" and self.state:
                self.to_disable_axis.add(ev["code"])

        if self.state:
            # If mode is pressed, forward all events
            return evs
        elif prev_state:
            # If prev_state, meaning the user released the mode or share button
            # turn off all buttons that were pressed during it
            for btn in self.to_disable_btn:
                out.append({"type": "button", "code": btn, "value": False})
            self.to_disable_btn = set()
            for axis in self.to_disable_axis:
                out.append({"type": "axis", "code": axis, "value": 0})
            self.to_disable_axis = set()
            return out
        else:
            # Otherwise, just return the standard buttons
            return out

    def consume(self, events: Sequence[Event]):
        return self.parent.consume(events)



import argparse


def evdev():
    from evdev import list_devices, InputDevice
    from time import sleep

    print("Available Devices with the Current Permissions")
    avail = list_devices()
    for i, d in enumerate(avail):
        print(f"{i + 1}:", InputDevice(d))

    print()
    sel = None
    while sel not in avail:
        sel = input("Enter device path (/dev/input/event# or #): ")
        try:
            sel = avail[int(sel) - 1]
        except Exception as e:
            pass

    print()
    d = InputDevice(sel)
    print(f"Selected device `{d}`.")
    print("Capabilities")
    print(d.capabilities(verbose=True))

    try:
        print("Attempting to grab device.")
        d.grab()
    except Exception as e:
        print("Could not grab device, error:")
        print(e)
    print()

    for ev in d.read_loop():
        print(ev)
        sleep(0.001)


def hidraw():
    from hhd.controller.lib.hid import enumerate_unique, Device
    from hhd.controller.lib.common import hexify
    from time import sleep, time

    print("Available Devices with the Current Permissions")
    avail = []
    for i, d in enumerate(enumerate_unique()):
        avail.append(d["path"])
        print(
            f"{i + 1}:",
            f"{str(d['path'])} {hexify(d['vendor_id'])}:{hexify(d['product_id'])}:"
            + f" Usage Page: 0x{hexify(d['usage_page'])} Usage: 0x{hexify(d['usage'])}"
            + f" Names '{d['manufacturer_string']}': '{d['product_string']}'",
        )

    print()
    sel = None
    while sel not in avail:
        sel = input("Enter device path (/dev/input/event# or #): ")
        try:
            sel = avail[int(sel) - 1]
        except Exception:
            sel = sel.encode()

    print()
    d = Device(path=sel)
    print(f"Selected device `{str(sel)}`.")

    start = time()
    for i in range(100000000):
        print(f"{i:6d}: {time() - start:7.4f}", d.read().hex())
        sleep(0.001)


def main():
    parser = argparse.ArgumentParser(
        prog="HHD: Handheld Daemon contribution helper scripts",
        description="Scripts to automate the capture of events, etc.",
    )
    parser.add_argument(
        "command",
        nargs="+",
        default=[],
        help="Supported commands: `evdev`, `hidraw`",
    )
    args = parser.parse_args()

    try:
        match c := args.command[0]:
            case "evdev":
                evdev()
            case "hidraw":
                hidraw()
            case _:
                print(f"Command `{c}` not supported.")
    except KeyboardInterrupt:
        pass




import logging
import os
import select
import subprocess
from threading import Event
from time import perf_counter, sleep
from typing import cast

import evdev
from evdev import ecodes as e

from hhd.utils import Context, expanduser

from .const import PowerButtonConfig

logger = logging.getLogger(__name__)

STEAM_PID = "~/.steam/steam.pid"
STEAM_EXE = "~/.steam/root/ubuntu12_32/steam"
STEAM_WAIT_DELAY = 0.5
LONG_PRESS_DELAY = 2.0


def B(b: str):
    return cast(int, getattr(evdev.ecodes, b))


def is_steam_gamescope_running(ctx: Context):
    pid = None
    try:
        with open(expanduser(STEAM_PID, ctx)) as f:
            pid = f.read().strip()

        steam_cmd_path = f"/proc/{pid}/cmdline"
        if not os.path.exists(steam_cmd_path):
            return False

        # Use this and line to determine if Steam is running in DeckUI mode.
        with open(steam_cmd_path, "rb") as f:
            steam_cmd = f.read()
        is_deck_ui = b"-gamepadui" in steam_cmd
        if not is_deck_ui:
            return False
    except Exception as e:
        return False
    return True


def run_steam_command(command: str, ctx: Context):
    global home_path
    try:
        if ctx.euid != ctx.uid:
            result = subprocess.run(
                [
                    "su",
                    ctx.name,
                    "-c",
                    f"{expanduser(STEAM_EXE, ctx)} -ifrunning {command}",
                ]
            )
        else:
            result = subprocess.run([expanduser(STEAM_EXE, ctx), "-ifrunning", command])

        return result.returncode == 0
    except Exception as e:
        logger.error(f"Received error when running steam command `{command}`\n{e}")
    return False


def register_power_buttons(b: PowerButtonConfig) -> list[evdev.InputDevice]:
    out = []
    for device in [evdev.InputDevice(path) for path in evdev.list_devices()]:
        capture = False
        for phys in b.phys:
            if str(device.phys).startswith(phys):
                capture = True
        if capture:
            device.grab()
            logger.info(f"Captured power button '{device.name}': '{device.phys}'")
            out.append(device)
    return out


def pick_closest_button(btns: list[evdev.InputDevice], cfg: PowerButtonConfig):
    for phys in cfg.phys:
        for b in btns:
            if str(b.phys).startswith(phys):
                return b

    if btns:
        return btns[0]
    return None


def register_hold_button(b: PowerButtonConfig) -> evdev.InputDevice | None:
    if not b.hold_phys or not b.hold_code:
        logger.error(
            f"Device configuration tuple does not contain required parameters:\n{b}"
        )
        return None

    for device in [evdev.InputDevice(path) for path in evdev.list_devices()]:
        for phys in b.hold_phys:
            if str(device.phys).startswith(phys):
                if b.hold_grab:
                    device.grab()
                logger.info(f"Captured hold keyboard '{device.name}': '{device.phys}'")
                return device
    return None


def run_steam_shortpress(perms: Context):
    return run_steam_command("steam://shortpowerpress", perms)


def run_steam_longpress(perms: Context):
    return run_steam_command("steam://longpowerpress", perms)


def power_button_run(cfg: PowerButtonConfig, ctx: Context, should_exit: Event):
    match cfg.type:
        case "hold_emitted":
            logger.info(
                f"Starting timer based powerbutton handler for device '{cfg.device}'."
            )
            power_button_timer(cfg, ctx, should_exit)
        case "hold_isa":
            logger.info(
                f"Starting isa keyboard powerbutton handler for device '{cfg.device}'."
            )
            power_button_isa(cfg, ctx, should_exit)
        case _:
            logger.error(f"Invalid type in config '{cfg.type}'. Exiting.")


def power_button_isa(cfg: PowerButtonConfig, perms: Context, should_exit: Event):
    press_dev = None
    press_devs = []
    hold_dev = None
    try:
        while not should_exit.is_set():
            # Initial check for steam
            if not is_steam_gamescope_running(perms):
                # Close devices
                if press_devs:
                    for d in press_devs:
                        d.close()
                    press_devs = []
                if press_dev:
                    press_dev.close()
                    press_dev = None
                if hold_dev:
                    hold_dev.close()
                    hold_dev = None
                logger.info(f"Waiting for steam to launch.")
                while not is_steam_gamescope_running(perms):
                    if should_exit.is_set():
                        return
                    sleep(STEAM_WAIT_DELAY)

            if not press_dev or not hold_dev:
                logger.info(f"Steam is running, hooking power button.")
                press_devs = register_power_buttons(cfg)
                press_dev = press_devs[0] if press_devs else None
                hold_dev = register_hold_button(cfg)
            if not press_dev:
                logger.error(f"Power button interfaces not found, disabling plugin.")
                return

            # Add timeout to release the button if steam exits.
            r = select.select(
                [press_dev.fd, hold_dev.fd] if hold_dev else [press_dev.fd],
                [],
                [],
                STEAM_WAIT_DELAY,
            )[0]

            if not r:
                continue
            fd = r[0]  # handle one button at a time

            # Handle button event
            issue_systemctl = False
            if fd == press_dev.fd:
                ev = press_dev.read_one()
                if ev.type == B("EV_KEY") and ev.code == B("KEY_POWER") and ev.value:
                    logger.info("Executing short press.")
                    issue_systemctl = not run_steam_shortpress(perms)
            elif hold_dev and fd == hold_dev.fd:
                ev = hold_dev.read_one()
                if ev.type == B("EV_KEY") and ev.code == cfg.hold_code and ev.value:
                    logger.info("Executing long press.")
                    issue_systemctl = not run_steam_longpress(perms)

            if issue_systemctl:
                logger.error(
                    "Power button action did not work. Calling `systemctl suspend`"
                )
                os.system("systemctl suspend")
    except KeyboardInterrupt:
        pass
    except Exception as e:
        logger.error(f"Received exception, exitting:\n{e}")


def power_button_timer(cfg: PowerButtonConfig, perms: Context, should_exit: Event):
    dev = None
    devs = []
    try:
        pressed_time = None
        while not should_exit.is_set():
            # Initial check for steam
            if not is_steam_gamescope_running(perms):
                # Close devices
                if devs:
                    for d in devs:
                        d.close()
                        devs = []
                    if dev:
                        dev.close()
                        dev = None
                logger.info(f"Waiting for steam to launch.")
                while not is_steam_gamescope_running(perms):
                    if should_exit.is_set():
                        return
                    sleep(STEAM_WAIT_DELAY)

            if not dev:
                logger.info(f"Steam is running, hooking power button.")
                devs = register_power_buttons(cfg)
                dev = pick_closest_button(devs, cfg)
            if not dev:
                logger.error(f"Power button not found, disabling plugin.")
                return

            # Add timeout to release the button if steam exits.
            delay = LONG_PRESS_DELAY if pressed_time else STEAM_WAIT_DELAY
            r = select.select([dev.fd], [], [], delay)[0]

            # Handle press logic
            if r:
                # Handle button event
                ev = dev.read_one()
                if ev.type == B("EV_KEY") and ev.code == B("KEY_POWER"):
                    curr_time = perf_counter()
                    if ev.value:
                        pressed_time = curr_time
                        press_type = "initial_press"
                    elif pressed_time:
                        if curr_time - pressed_time > LONG_PRESS_DELAY:
                            press_type = "long_press"
                        else:
                            press_type = "short_press"
                        pressed_time = None
                    else:
                        press_type = "release_without_press"
                else:
                    press_type = "no_press"
            elif pressed_time:
                # Button was pressed but we hit a timeout, that means
                # it is a long press
                press_type = "long_press"
                pressed_time = None
            else:
                # Otherwise, no press
                press_type = "no_press"

            issue_systemctl = False
            match press_type:
                case "long_press":
                    logger.info("Executing long press.")
                    issue_systemctl = not run_steam_longpress(perms)
                case "short_press":
                    logger.info("Executing short press.")
                    issue_systemctl = not run_steam_shortpress(perms)
                case "initial_press":
                    logger.info("Power button pressed down.")
                case "release_without_press":
                    logger.error("Button released without being pressed.")

            if issue_systemctl:
                logger.error(
                    "Power button action did not work. Calling `systemctl suspend`"
                )
                os.system("systemctl suspend")
    except KeyboardInterrupt:
        pass
    except Exception as e:
        logger.error(f"Received exception, exitting:\n{e}")
    finally:
        if dev:
            dev.close()
        if devs:
            for d in devs:
                d.close()



from hhd.controller import Axis, Button, Configuration
from hhd.controller.physical.evdev import B, to_map
from hhd.controller.physical.hidraw import AM, BM, CM

LGO_TOUCHPAD_BUTTON_MAP: dict[int, Button] = to_map(
    {
        "touchpad_touch": [B("BTN_TOOL_FINGER")],  # also BTN_TOUCH
        "touchpad_right": [B("BTN_TOOL_DOUBLETAP")],
    }
)

LGO_TOUCHPAD_AXIS_MAP: dict[int, Axis] = to_map(
    {
        "touchpad_x": [B("ABS_X")],  # also ABS_MT_POSITION_X
        "touchpad_y": [B("ABS_Y")],  # also ABS_MT_POSITION_Y
    }
)

LGO_RAW_INTERFACE_BTN_ESSENTIALS: dict[int | None, dict[Button, BM]] = {
    0x74: {
        # Misc
        "mode": BM((18 << 3)),
        "share": BM((18 << 3) + 1),
        # Back buttons
        "extra_l1": BM((20 << 3)),
        "extra_l2": BM((20 << 3) + 1),
        "extra_r1": BM((20 << 3) + 2),
        "extra_r2": BM((20 << 3) + 5),
        "extra_r3": BM((20 << 3) + 4),
    }
}


LGO_RAW_INTERFACE_BTN_MAP: dict[int | None, dict[Button, BM]] = {
    0x74: {
        # Misc
        "mode": BM((18 << 3)),
        "share": BM((18 << 3) + 1),
        # Sticks
        "ls": BM((18 << 3) + 2),
        "rs": BM((18 << 3) + 3),
        # D-PAD
        "dpad_up": BM((18 << 3) + 4),
        "dpad_down": BM((18 << 3) + 5),
        "dpad_left": BM((18 << 3) + 6),
        "dpad_right": BM((18 << 3) + 7),
        # Thumbpad
        "a": BM((19 << 3) + 0),
        "b": BM((19 << 3) + 1),
        "x": BM((19 << 3) + 2),
        "y": BM((19 << 3) + 3),
        # Bumpers
        "lb": BM((19 << 3) + 4),
        "lt": BM((19 << 3) + 5),
        "rb": BM((19 << 3) + 6),
        "rt": BM((19 << 3) + 7),
        # Back buttons
        "extra_l1": BM((20 << 3)),
        "extra_l2": BM((20 << 3) + 1),
        "extra_r1": BM((20 << 3) + 2),
        "extra_r2": BM((20 << 3) + 5),
        "extra_r3": BM((20 << 3) + 4),
        # Select
        "start": BM((20 << 3) + 7),
        "select": BM((20 << 3) + 6),
        # Mouse
        "btn_middle": BM((21 << 3)),
    }
}


LGO_RAW_INTERFACE_AXIS_MAP: dict[int | None, dict[Axis, AM]] = {
    0x74: {
        "ls_x": AM(14 << 3, "m8"),
        "ls_y": AM(15 << 3, "m8"),
        "rs_x": AM(16 << 3, "m8"),
        "rs_y": AM(17 << 3, "m8"),
        "rt": AM(22 << 3, "u8"),
        "lt": AM(23 << 3, "u8"),
        # "mouse_wheel": AM(25 << 3, "m8", scale=1), # TODO: Fix weird behavior
        # "touchpad_x": AM(26 << 3, "u16"),
        # "touchpad_y": AM(28 << 3, "u16"),
        # "left_gyro_x": AM(30 << 3, "m8"),
        # "left_gyro_y": AM(31 << 3, "m8"),
        # "right_gyro_x": AM(32 << 3, "m8"),
        # "right_gyro_y": AM(33 << 3, "m8"),
    }
}

LGO_RAW_INTERFACE_CONFIG_MAP: dict[int | None, dict[Configuration, CM]] = {
    0x74: {
        "battery_left": CM(5 << 3, "u8", scale=1, bounds=(0, 100)),
        "battery_right": CM(7 << 3, "u8", scale=1, bounds=(0, 100)),
        "is_connected_left": CM((10 << 3) + 7, "bit"),
        "is_connected_right": CM((11 << 3) + 7, "bit"),
        "is_attached_left": CM((12 << 3) + 7, "bit", flipped=True),
        "is_attached_right": CM((13 << 3) + 7, "bit", flipped=True),
    }
}



from typing import Literal


AbsAxis = Literal[
    # Sticks
    # Values should range from -1 to 1
    "ls_x",
    "ls_y",
    "rs_x",
    "rs_y",
    # Triggers
    # Values should range from 0 to 1
    "lt",
    "rt",
    # Hat, implemented as axis. Either -1, 0, or 1
    "hat_x",
    "hat_y",
    # Accelerometer
    # Values should be in m2/s
    "accel_x",
    "accel_y",
    "accel_z",
    "accel_ts",
    # Gyroscope
    # Values should be in deg/s
    "gyro_x",
    "gyro_y",
    "gyro_z",
    "gyro_ts",
    # Touchpad
    # Both width and height should go from [0, 1]. Aspect ratio is a setting.
    # It is up to the device whether to stretch, crop and how to crop (either
    # crop part of the input or part of itself)
    "touchpad_x",
    "touchpad_y",
    # Sidedness
    # Left
    "left_accel_x",
    "left_accel_y",
    "left_accel_z",
    "left_accel_ts",
    "left_gyro_x",
    "left_gyro_y",
    "left_gyro_z",
    "left_gyro_ts",
    "left_touchpad_x",
    "left_touchpad_y",
    # Right
    "right_accel_x",
    "right_accel_y",
    "right_accel_z",
    "right_accel_ts",
    "right_gyro_x",
    "right_gyro_y",
    "right_gyro_z",
    "right_gyro_ts",
    "right_touchpad_x",
    "right_touchpad_y",
]

RelAxis = Literal["mouse_x", "mouse_y", "mouse_wheel", "mouse_wheel_hires"]

GamepadButton = Literal[
    # Thumbpad
    "a",
    "b",
    "x",
    "y",
    # D-PAD (avail as both axis and buttons)
    "dpad_up",
    "dpad_down",
    "dpad_left",
    "dpad_right",
    # Sticks
    "ls",
    "rs",
    # Bumpers
    "lb",
    "rb",
    # Triggers
    "lt",
    "rt",
    # Back buttons
    "extra_l1",
    "extra_l2",
    "extra_l3",
    "extra_r1",
    "extra_r2",
    "extra_r3",
    # Select
    "start",
    "select",
    # Misc
    "mode",
    "share",
    # Touchpad
    "touchpad_touch",
    "touchpad_left",
    "touchpad_right",
]

MouseButton = Literal["btn_left", "btn_right", "btn_middle", "btn_side", "btn_extra"]

KeyboardButton = Literal[
    "key_esc",  # 1
    "key_enter",  # 28
    # Modifiers
    "key_leftctrl",  # 29
    "key_leftshift",  # 42
    "key_leftalt",  # 56
    "key_rightctrl",  # 97
    "key_rightshift",  # 54
    "key_rightalt",  # 100
    "key_leftmeta",  # 125
    "key_rightmeta",  # 126
    # Special Keys
    "key_capslock",  # 58
    "key_numlock",  # 69
    "key_scrolllock",  # 70
    "key_sysrq",  # 99
    # Symbols
    "key_minus",  # 12
    "key_equal",  # 13
    "key_backspace",  # 14
    "key_tab",  # 15
    "key_leftbrace",  # 26
    "key_rightbrace",  # 27
    "key_space",  # 57
    "key_up",  # 103
    "key_left",  # 105
    "key_right",  # 106
    "key_down",  # 108
    "key_home",  # 102
    "key_end",  # 107
    "key_pageup",  # 104
    "key_pagedown",  # 109
    "key_insert",  # 110
    "key_delete",  # 111
    "key_semicolon",  # 39
    "key_apostrophe",  # 40
    "key_grave",  # 41
    "key_backslash",  # 43
    "key_comma",  # 51
    "key_dot",  # 52
    "key_slash",  # 53
    "key_102nd",  # 86
    "key_ro",  # 89
    "key_power",  # 116
    "key_compose",  # 127
    "key_stop",  # 128
    "key_again",  # 129
    "key_props",  # 130
    "key_undo",  # 131
    "key_front",  # 132
    "key_copy",  # 133
    "key_open",  # 134
    "key_paste",  # 135
    "key_cut",  # 137
    "key_find",  # 136
    "key_help",  # 138
    "key_calc",  # 140
    "key_sleep",  # 142
    "key_www",  # 150
    "key_screenlock",  # 152
    "key_back",  # 158
    "key_refresh",  # 173
    "key_edit",  # 176
    "key_scrollup",  # 177
    "key_scrolldown",  # 178
    # Numbers
    "key_1",  # 2
    "key_2",  # 3
    "key_3",  # 4
    "key_4",  # 5
    "key_5",  # 6
    "key_6",  # 7
    "key_7",  # 8
    "key_8",  # 9
    "key_9",  # 10
    "key_0",  # 11
    # Letters
    "key_a",  # 30
    "key_b",  # 48
    "key_c",  # 46
    "key_d",  # 32
    "key_e",  # 18
    "key_f",  # 33
    "key_g",  # 34
    "key_h",  # 35
    "key_i",  # 23
    "key_j",  # 36
    "key_k",  # 37
    "key_l",  # 38
    "key_m",  # 50
    "key_n",  # 49
    "key_o",  # 24
    "key_p",  # 25
    "key_q",  # 16
    "key_r",  # 19
    "key_s",  # 31
    "key_t",  # 20
    "key_u",  # 22
    "key_v",  # 47
    "key_w",  # 17
    "key_x",  # 45
    "key_y",  # 21
    "key_z",  # 44
    # Keypad Keys
    "key_kpasterisk",  # 55
    "key_kpminus",  # 74
    "key_kpplus",  # 78
    "key_kpdot",  # 83
    "key_kpjpcomma",  # 95
    "key_kpenter",  # 96
    "key_kpslash",  # 98
    "key_kpequal",  # 117
    "key_kpcomma",  # 121
    "key_kpleftparen",  # 179
    "key_kprightparen",  # 180
    # Keypad Numbers
    "key_kp0",  # 82
    "key_kp1",  # 79
    "key_kp2",  # 80
    "key_kp3",  # 81
    "key_kp4",  # 75
    "key_kp5",  # 76
    "key_kp6",  # 77
    "key_kp7",  # 71
    "key_kp8",  # 72
    "key_kp9",  # 73
    # Function keys
    "key_f1",  # 59
    "key_f2",  # 60
    "key_f3",  # 61
    "key_f4",  # 62
    "key_f5",  # 63
    "key_f6",  # 64
    "key_f7",  # 65
    "key_f8",  # 66
    "key_f9",  # 67
    "key_f11",  # 87
    "key_f12",  # 88
    "key_f10",  # 68
    "key_f13",  # 183
    "key_f14",  # 184
    "key_f15",  # 185
    "key_f16",  # 186
    "key_f17",  # 187
    "key_f18",  # 188
    "key_f19",  # 189
    "key_f20",  # 190
    "key_f21",  # 191
    "key_f22",  # 192
    "key_f23",  # 193
    "key_f24",  # 194
    # Media Keys
    "key_playpause",  # 164
    "key_pause",  # 119
    "key_mute",  # 113
    "key_stopcd",  # 166
    "key_forward",  # 159
    "key_ejectcd",  # 161
    "key_nextsong",  # 163
    "key_previoussong",  # 165
    "key_volumedown",  # 114
    "key_volumeup",  # 115
    # Language specific
    "key_katakana",  # 90
    "key_hiragana",  # 91
    "key_henkan",  # 92
    "key_katakanahiragana",  # 93
    "key_muhenkan",  # 94
    "key_zenkakuhankaku",  # 85
    "key_hanguel",  # 122
    "key_hanja",  # 123
    "key_yen",  # 124
    # ?
    "key_unknown",  # 240,
    # Prog for the ally
    "key_prog1",
    "key_prog2",
]

Axis = AbsAxis | RelAxis
Button = GamepadButton | KeyboardButton | MouseButton

Configuration = Literal[
    # Misc
    "led_mute",  # binary
    "player",
    # Set the aspect ratio of the touchpad used
    # width / height
    "touchpad_aspect_ratio",
    "battery",
    "is_connected",
    "is_attached",
    "battery_left",
    "battery_right",
    "is_connected_left",
    "is_connected_right",
    "is_attached_left",
    "is_attached_right",
]



from .conf import Config
from .plugin import HHDAutodetect, HHDPlugin, Context, Emitter, Event
from .settings import HHDSettings
from .utils import get_relative_fn, load_relative_yaml
from .outputs import get_outputs_config, get_outputs
from .inputs import get_touchpad_config, get_gyro_config, get_gyro_state, gen_gyro_state


__all__ = [
    "Config",
    "HHDSettings",
    "HHDAutodetect",
    "HHDPlugin",
    "get_relative_fn",
    "load_relative_yaml",
    "Emitter",
    "Event",
    "Context",
    "get_outputs_config",
    "get_outputs",
    "get_touchpad_config",
    "get_gyro_config",
    "get_gyro_state",
    "gen_gyro_state",
]



from hhd.controller import Axis, Button, Configuration
from hhd.controller.physical.evdev import B, to_map

GPD_TOUCHPAD_BUTTON_MAP: dict[int, Button] = to_map(
    {
        "touchpad_touch": [B("BTN_TOOL_FINGER")],  # also BTN_TOUCH
        "touchpad_right": [B("BTN_TOOL_DOUBLETAP"), B("BTN_RIGHT")],
        "touchpad_left": [B("BTN_MOUSE")],
    }
)

GPD_TOUCHPAD_AXIS_MAP: dict[int, Axis] = to_map(
    {
        "touchpad_x": [B("ABS_X")],  # also ABS_MT_POSITION_X
        "touchpad_y": [B("ABS_Y")],  # also ABS_MT_POSITION_Y
    }
)

GPD_WIN_DEFAULT_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("accel_z", "accel", 1, 3),
    "accel_y": ("accel_x", "accel", 1, 3),
    "accel_z": ("accel_y", "accel", 1, 3),
    "anglvel_x": ("gyro_x", "anglvel", 1, None),
    "anglvel_y": ("gyro_z", "anglvel", -1, None),
    "anglvel_z": ("gyro_y", "anglvel", -1, None),
    "timestamp": ("gyro_ts", None, 1, None),
}

GPD_WIN_MAX_2_2023_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("accel_z", "accel", 1, 3),
    "accel_y": ("accel_x", "accel", 1, 3),
    "accel_z": ("accel_y", "accel", 1, 3),
    "anglvel_x": ("gyro_z", "anglvel", 1, None),
    "anglvel_y": ("gyro_x", "anglvel", -1, None),
    "anglvel_z": ("gyro_y", "anglvel", -1, None),
    "timestamp": ("gyro_ts", None, 1, None),
}



import logging
import re
import select
import time
from threading import Event as TEvent
from typing import Sequence

import evdev

from hhd.controller import Axis, Event, Multiplexer, can_read
from hhd.controller.base import Event, TouchpadAction
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev
from hhd.controller.physical.hidraw import GenericGamepadHidraw
from hhd.controller.physical.imu import CombinedImu, HrtimerTrigger
from hhd.plugins import Config, Context, Emitter, get_outputs, get_gyro_state

from .const import (
    GPD_TOUCHPAD_AXIS_MAP,
    GPD_TOUCHPAD_BUTTON_MAP,
    GPD_WIN_DEFAULT_MAPPINGS,
)

ERROR_DELAY = 1
SELECT_TIMEOUT = 1

logger = logging.getLogger(__name__)

GPD_WIN_4_VID = 0x2F24
GPD_WIN_4_PID = 0x0135
GAMEPAD_VID = 0x045E
GAMEPAD_PID = 0x028E

TOUCHPAD_VID = 0x093A
TOUCHPAD_PID = 0x0255
TOUCHPAD_VID_2 = 0x0911
TOUCHPAD_PID_2 = 0x5288

BACK_BUTTON_DELAY = 0.1

# /dev/input/event17 Microsoft X-Box 360 pad usb-0000:73:00.3-4.1/input0
# bus: 0003, vendor 045e, product 028e, version 0101

# back buttons
# /dev/input/event15   Mouse for Windows usb-0000:73:00.3-4.2/input1
# bus: 0003, vendor 2f24, product 0135, version 0110

# physical keyboard
# /dev/input/event13   Mouse for Windows usb-0000:73:00.3-4.2/input0
# bus: 0003, vendor 2f24, product 0135, version 0110

# hidraw back buttons  {'path': b'/dev/hidraw6',
#    'vendor_id': 12068, 'product_id': 309, 'serial_number': '',
#    'release_number': 256, 'manufacturer_string': ' ',
#    'product_string': 'Mouse for Windows',
#    'usage_page': 1, 'usage': 6, 'interface_number': 1},


class GpdWin4Hidraw(GenericGamepadHidraw):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def open(self) -> Sequence[int]:
        self.left_pressed = None
        self.right_pressed = None
        self.last_pressed = None
        self.clear_ts = None

        self.queue: list[tuple[Event, float]] = []
        return super().open()

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        # If we can not read return
        if not self.fd or not self.dev:
            return []

        # Process events
        curr = time.perf_counter()
        out: Sequence[Event] = []

        # Read new events
        left_pressed = None
        right_pressed = None
        while can_read(self.fd):
            rep = self.dev.read(self.report_size)

            # l4 = 0x46
            # r4 = 0x48
            # both = l4 + r4
            # when both l4/r4 held, rep[2] and rep[3] will both be active
            #   they will be the same known values for l4 and r4
            #   but the order is not guaranteed to be consistent
            check = rep[2] + rep[3]
            match check:
                case 0x46:
                    # action = "left/l4"
                    left_pressed = True
                    self.last_pressed = "left"
                    self.clear_ts = None
                case 0x48:
                    # action = "right/r4"
                    right_pressed = True
                    self.last_pressed = "right"
                    self.clear_ts = None
                case 0x8E:
                    # both l4 and r4 are being pressed
                    left_pressed = True
                    right_pressed = True
                    self.clear_ts = None
                case _:  # 0x00:
                    # This occurs only when one button is pressed
                    # So in case both are remove one
                    if self.last_pressed == "right" and self.left_pressed:
                        left_pressed = False
                    if self.last_pressed == "left" and self.right_pressed:
                        right_pressed = False
                    self.clear_ts = curr + BACK_BUTTON_DELAY

        if self.clear_ts and self.clear_ts < curr:
            # Reset after timeout
            if self.left_pressed:
                out.append({"type": "button", "code": "extra_l1", "value": False})
                self.left_pressed = False
            if self.right_pressed:
                out.append({"type": "button", "code": "extra_r1", "value": False})
                self.right_pressed = False
            self.clear_ts = None
        else:
            # If no timeout, update
            # Left, right will be none if no events were received
            # If they were, they will be true/false
            # If that conflicts with the saved values, send events.
            if left_pressed is not None and self.left_pressed != left_pressed:
                out.append(
                    {"type": "button", "code": "extra_l1", "value": left_pressed}
                )
                self.left_pressed = left_pressed

            if right_pressed is not None and self.right_pressed != right_pressed:
                out.append(
                    {"type": "button", "code": "extra_r1", "value": right_pressed}
                )
                self.right_pressed = right_pressed
        return out


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
):
    first = True
    while not should_exit.is_set():
        if conf["controller_mode.mode"].to(str) == "disabled":
            time.sleep(ERROR_DELAY)
            continue

        found_gamepad = False
        try:
            for d in evdev.list_devices():
                dev = evdev.InputDevice(d)
                if dev.info.vendor == GAMEPAD_VID and dev.info.product == GAMEPAD_PID:
                    found_gamepad = True
                    break
        except Exception:
            logger.warning("Failed finding device, skipping check.")
            found_gamepad = True

        if not found_gamepad:
            if first:
                logger.info("Controller in Mouse mode. Waiting...")
            time.sleep(ERROR_DELAY)
            first = False
            continue

        try:
            logger.info("Launching emulated controller.")
            updated.clear()
            controller_loop(conf.copy(), should_exit, updated, dconf)
        except Exception as e:
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {ERROR_DELAY}s."
            )
            first = True
            # Raise exception
            if conf.get("debug", False):
                raise e
            time.sleep(ERROR_DELAY)


def controller_loop(conf: Config, should_exit: TEvent, updated: TEvent, dconf: dict):
    debug = conf.get("debug", False)
    has_touchpad = dconf.get("touchpad", False)

    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        conf["touchpad"] if has_touchpad else None,
        conf["imu"].to(bool),
    )

    # Imu
    d_imu = CombinedImu(
        conf["imu_hz"].to(int),
        get_gyro_state(
            conf["imu_axis"], dconf.get("mapping", GPD_WIN_DEFAULT_MAPPINGS)
        ),
    )
    d_timer = HrtimerTrigger(conf["imu_hz"].to(int), [HrtimerTrigger.IMU_NAMES])

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[GAMEPAD_VID],
        pid=[GAMEPAD_PID],
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
    )

    # "PNP0C50:00 0911:5288 Touchpad" on Win Max 2 2023
    # "PNP0C50:00 093A:0255 Touchpad" on Win Mini
    d_touch = GenericGamepadEvdev(
        vid=[TOUCHPAD_VID, TOUCHPAD_VID_2],
        pid=[TOUCHPAD_PID, TOUCHPAD_PID_2],
        name=[re.compile(".+Touchpad")],
        capabilities={EC("EV_KEY"): [EC("BTN_MOUSE")]},
        btn_map=GPD_TOUCHPAD_BUTTON_MAP,
        axis_map=GPD_TOUCHPAD_AXIS_MAP,
        aspect_ratio=1.333,
        required=False,
    )

    # Vendor
    d_vend = GpdWin4Hidraw(
        vid=[GPD_WIN_4_VID],
        pid=[GPD_WIN_4_PID],
        usage_page=[0x0001],
        usage=[0x0006],
        required=True,
    )

    d_kbd_1 = GenericGamepadEvdev(
        vid=[GPD_WIN_4_VID],
        pid=[GPD_WIN_4_PID],
        # TODO: Verify capability check does not cause regressions
        capabilities={EC("EV_KEY"): [EC("KEY_SYSRQ"), EC("KEY_PAUSE")]},
        required=False,
        grab=True,
        # btn_map={EC("KEY_SYSRQ"): "extra_l1", EC("KEY_PAUSE"): "extra_r1"},
    )

    match conf["l4r4_to_qam"].to(str):
        case "l4":
            qam_button = "extra_l1"
        case "r4":
            qam_button = "extra_r1"
        case _:
            qam_button = None

    if has_touchpad:
        touch_actions = (
            conf["touchpad.controller"]
            if conf["touchpad.mode"].to(TouchpadAction) == "controller"
            else conf["touchpad.emulation"]
        )

        multiplexer = Multiplexer(
            trigger="analog_to_discrete",
            dpad="analog_to_discrete",
            touchpad_short=touch_actions["short"].to(TouchpadAction),
            touchpad_hold=touch_actions["hold"].to(TouchpadAction),
            nintendo_mode=conf["nintendo_mode"].to(bool),
            qam_button=qam_button,
        )
    else:
        multiplexer = Multiplexer(
            trigger="analog_to_discrete",
            dpad="analog_to_discrete",
            nintendo_mode=conf["nintendo_mode"].to(bool),
            qam_button=qam_button,
        )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    if conf["imu"].to(bool):
        REPORT_FREQ_MAX = max(REPORT_FREQ_MAX, conf["imu_hz"].to(float))

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        d_vend.open()
        prepare(d_xinput)
        if conf.get("imu", False):
            start_imu = True
            if dconf.get("hrtimer", False):
                start_imu = d_timer.open()
            if start_imu:
                prepare(d_imu)
        if has_touchpad and d_params["uses_touch"]:
            prepare(d_touch)
        prepare(d_kbd_1)
        for d in d_producers:
            prepare(d)

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))
            evs.extend(d_vend.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                # d_vend.consume(evs)
                d_xinput.consume(evs)

            for d in d_outs:
                d.consume(evs)

            # If unbounded, the total number of events per second is the sum of all
            # events generated by the producers.
            # For Legion go, that would be 100 + 100 + 500 + 30 = 730
            # Since the controllers of the legion go only update at 500hz, this is
            # wasteful.
            # By setting a target refresh rate for the report and sleeping at the
            # end, we ensure that even if multiple fds become ready close to each other
            # they are combined to the same report, limiting resource use.
            # Ideally, this rate is smaller than the report rate of the hardware controller
            # to ensure there is always a report from that ready during refresh
            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        d_vend.close(True)
        d_timer.close()
        for d in reversed(devs):
            d.close(True)



import logging
from typing import cast, Sequence, NamedTuple

import evdev
from evdev import AbsInfo

from hhd.controller.const import Axis, Button


HHD_VID = 0x5335
HHD_PID_GAMEPAD = 0x01
HHD_PID_KEYBOARD = 0x02
HHD_PID_MOUSE = 0x03
HHD_PID_TOUCHPAD = 0x04
HHD_PID_MOTION = 0x11
HHD_PID_VENDOR = 0x7000


def B(b: str | Sequence[str], num: int | None = None):
    if num is not None:
        return num
    assert b, f"No value provided."
    if not isinstance(b, str):
        b = b[0]
    return cast(int, getattr(evdev.ecodes, b))


class AX(NamedTuple):
    id: int
    scale: float = 1
    offset: float = 0
    bounds: tuple[int, int] | None = None


logger = logging.getLogger(__name__)

GAMEPAD_BTN_CAPABILITIES = {
    B("EV_KEY"): [
        B("BTN_TL"),
        B("BTN_TR"),
        B("BTN_SELECT"),
        B("BTN_START"),
        B("BTN_MODE"),
        B("BTN_THUMBL"),
        B("BTN_THUMBR"),
        B("BTN_A"),
        B("BTN_B"),
        B("BTN_X"),
        B("BTN_Y"),
        B("BTN_TRIGGER_HAPPY1"),
        B("BTN_TRIGGER_HAPPY2"),
        B("BTN_TRIGGER_HAPPY3"),
        B("BTN_TRIGGER_HAPPY4"),
        B("BTN_TRIGGER_HAPPY5"),
        B("BTN_TRIGGER_HAPPY6"),
    ]
}

GAMEPAD_CAPABILITIES = {
    # B("EV_SYN", 0): [
    #     B("SYN_REPORT", 0),
    #     B("SYN_CONFIG", 1),
    #     B("SYN_DROPPED", 3),
    #     B("?", 21),
    # ],
    B("EV_KEY", 1): [
        B(["BTN_A", "BTN_GAMEPAD", "BTN_SOUTH"], 304),
        B(["BTN_B", "BTN_EAST"], 305),
        B(["BTN_NORTH", "BTN_X"], 307),
        B(["BTN_WEST", "BTN_Y"], 308),
        B("BTN_TL", 310),
        B("BTN_TR", 311),
        B("BTN_SELECT", 314),
        B("BTN_START", 315),
        B("BTN_MODE", 316),
        B("BTN_THUMBL", 317),
        B("BTN_THUMBR", 318),
        B("BTN_TRIGGER_HAPPY1"),
        B("BTN_TRIGGER_HAPPY2"),
        B("BTN_TRIGGER_HAPPY3"),
        B("BTN_TRIGGER_HAPPY4"),
        B("BTN_TRIGGER_HAPPY5"),
        B("BTN_TRIGGER_HAPPY6"),
        B("BTN_TRIGGER_HAPPY7"),
        B("BTN_TRIGGER_HAPPY8"),
        B("BTN_TRIGGER_HAPPY9"),
        B("BTN_TRIGGER_HAPPY10"),
        B("BTN_TRIGGER_HAPPY11"),
        B("BTN_TRIGGER_HAPPY12"),
        B("BTN_TRIGGER_HAPPY13"),
        B("BTN_TRIGGER_HAPPY14"),
        B("BTN_TRIGGER_HAPPY15"),
        B("BTN_TRIGGER_HAPPY16"),
        B("BTN_TRIGGER_HAPPY17"),
        B("BTN_TRIGGER_HAPPY18"),
        B("BTN_TRIGGER_HAPPY19"),
        B("BTN_TRIGGER_HAPPY20"),
    ],
    B("EV_ABS", 3): [
        (
            B("ABS_X", 0),
            AbsInfo(value=0, min=-32768, max=32767, fuzz=16, flat=128, resolution=0),
        ),
        (
            B("ABS_Y", 1),
            AbsInfo(value=0, min=-32768, max=32767, fuzz=16, flat=128, resolution=0),
        ),
        (B("ABS_Z", 2), AbsInfo(value=0, min=0, max=255, fuzz=0, flat=0, resolution=0)),
        (
            B("ABS_RX", 3),
            AbsInfo(value=0, min=-32768, max=32767, fuzz=16, flat=128, resolution=0),
        ),
        (
            B("ABS_RY", 4),
            AbsInfo(value=0, min=-32768, max=32767, fuzz=16, flat=128, resolution=0),
        ),
        (
            B("ABS_RZ", 5),
            AbsInfo(value=0, min=0, max=255, fuzz=0, flat=0, resolution=0),
        ),
        (
            B("ABS_HAT0X", 16),
            AbsInfo(value=0, min=-1, max=1, fuzz=0, flat=0, resolution=0),
        ),
        (
            B("ABS_HAT0Y", 17),
            AbsInfo(value=0, min=-1, max=1, fuzz=0, flat=0, resolution=0),
        ),
    ],
    B("EV_FF", 21): [
        B(["FF_EFFECT_MIN", "FF_RUMBLE"], 80),
        B("FF_PERIODIC", 81),
        B(["FF_SQUARE", "FF_WAVEFORM_MIN"], 88),
        B("FF_TRIANGLE", 89),
        B("FF_SINE", 90),
        B(["FF_GAIN", "FF_MAX_EFFECTS"], 96),
    ],
}

MOTION_CAPABILITIES = {
    # B("EV_SYN", 0): [B("SYN_REPORT", 0), B("SYN_DROPPED", 3), B("?", 4)],
    B("EV_ABS", 3): [
        (
            B("ABS_X", 0),
            AbsInfo(value=0, min=-32768, max=32768, fuzz=16, flat=0, resolution=8192),
        ),
        (
            B("ABS_Y", 1),
            AbsInfo(value=0, min=-32768, max=32768, fuzz=16, flat=0, resolution=8192),
        ),
        (
            B("ABS_Z", 2),
            AbsInfo(value=0, min=-32768, max=32768, fuzz=16, flat=0, resolution=8192),
        ),
        (
            B("ABS_RX", 3),
            AbsInfo(
                value=0, min=-2097152, max=2097152, fuzz=16, flat=0, resolution=1024
            ),
        ),
        (
            B("ABS_RY", 4),
            AbsInfo(
                value=0, min=-2097152, max=2097152, fuzz=16, flat=0, resolution=1024
            ),
        ),
        (
            B("ABS_RZ", 5),
            AbsInfo(
                value=0, min=-2097152, max=2097152, fuzz=16, flat=0, resolution=1024
            ),
        ),
    ],
    B("EV_MSC", 4): [B("MSC_TIMESTAMP", 5)],
}

KEYBOARD_CAPABILITIES = {
    # B("EV_SYN", 0): [
    #     B("SYN_REPORT", 0),
    #     B("SYN_CONFIG", 1),
    #     B("?", 4),
    #     B("?", 17),
    #     B("?", 20),
    # ],
    B("EV_KEY", 1): [
        B("KEY_ESC", 1),
        B("KEY_1", 2),
        B("KEY_2", 3),
        B("KEY_3", 4),
        B("KEY_4", 5),
        B("KEY_5", 6),
        B("KEY_6", 7),
        B("KEY_7", 8),
        B("KEY_8", 9),
        B("KEY_9", 10),
        B("KEY_0", 11),
        B("KEY_MINUS", 12),
        B("KEY_EQUAL", 13),
        B("KEY_BACKSPACE", 14),
        B("KEY_TAB", 15),
        B("KEY_Q", 16),
        B("KEY_W", 17),
        B("KEY_E", 18),
        B("KEY_R", 19),
        B("KEY_T", 20),
        B("KEY_Y", 21),
        B("KEY_U", 22),
        B("KEY_I", 23),
        B("KEY_O", 24),
        B("KEY_P", 25),
        B("KEY_LEFTBRACE", 26),
        B("KEY_RIGHTBRACE", 27),
        B("KEY_ENTER", 28),
        B("KEY_LEFTCTRL", 29),
        B("KEY_A", 30),
        B("KEY_S", 31),
        B("KEY_D", 32),
        B("KEY_F", 33),
        B("KEY_G", 34),
        B("KEY_H", 35),
        B("KEY_J", 36),
        B("KEY_K", 37),
        B("KEY_L", 38),
        B("KEY_SEMICOLON", 39),
        B("KEY_APOSTROPHE", 40),
        B("KEY_GRAVE", 41),
        B("KEY_LEFTSHIFT", 42),
        B("KEY_BACKSLASH", 43),
        B("KEY_Z", 44),
        B("KEY_X", 45),
        B("KEY_C", 46),
        B("KEY_V", 47),
        B("KEY_B", 48),
        B("KEY_N", 49),
        B("KEY_M", 50),
        B("KEY_COMMA", 51),
        B("KEY_DOT", 52),
        B("KEY_SLASH", 53),
        B("KEY_RIGHTSHIFT", 54),
        B("KEY_KPASTERISK", 55),
        B("KEY_LEFTALT", 56),
        B("KEY_SPACE", 57),
        B("KEY_CAPSLOCK", 58),
        B("KEY_F1", 59),
        B("KEY_F2", 60),
        B("KEY_F3", 61),
        B("KEY_F4", 62),
        B("KEY_F5", 63),
        B("KEY_F6", 64),
        B("KEY_F7", 65),
        B("KEY_F8", 66),
        B("KEY_F9", 67),
        B("KEY_F10", 68),
        B("KEY_NUMLOCK", 69),
        B("KEY_SCROLLLOCK", 70),
        B("KEY_KP7", 71),
        B("KEY_KP8", 72),
        B("KEY_KP9", 73),
        B("KEY_KPMINUS", 74),
        B("KEY_KP4", 75),
        B("KEY_KP5", 76),
        B("KEY_KP6", 77),
        B("KEY_KPPLUS", 78),
        B("KEY_KP1", 79),
        B("KEY_KP2", 80),
        B("KEY_KP3", 81),
        B("KEY_KP0", 82),
        B("KEY_KPDOT", 83),
        B("KEY_ZENKAKUHANKAKU", 85),
        B("KEY_102ND", 86),
        B("KEY_F11", 87),
        B("KEY_F12", 88),
        B("KEY_RO", 89),
        B("KEY_KATAKANA", 90),
        B("KEY_HIRAGANA", 91),
        B("KEY_HENKAN", 92),
        B("KEY_KATAKANAHIRAGANA", 93),
        B("KEY_MUHENKAN", 94),
        B("KEY_KPJPCOMMA", 95),
        B("KEY_KPENTER", 96),
        B("KEY_RIGHTCTRL", 97),
        B("KEY_KPSLASH", 98),
        B("KEY_SYSRQ", 99),
        B("KEY_RIGHTALT", 100),
        B("KEY_HOME", 102),
        B("KEY_UP", 103),
        B("KEY_PAGEUP", 104),
        B("KEY_LEFT", 105),
        B("KEY_RIGHT", 106),
        B("KEY_END", 107),
        B("KEY_DOWN", 108),
        B("KEY_PAGEDOWN", 109),
        B("KEY_INSERT", 110),
        B("KEY_DELETE", 111),
        B(["KEY_MIN_INTERESTING", "KEY_MUTE"], 113),
        B("KEY_VOLUMEDOWN", 114),
        B("KEY_VOLUMEUP", 115),
        B("KEY_POWER", 116),
        B("KEY_KPEQUAL", 117),
        B("KEY_PAUSE", 119),
        B("KEY_KPCOMMA", 121),
        B(["KEY_HANGEUL", "KEY_HANGUEL"], 122),
        B("KEY_HANJA", 123),
        B("KEY_YEN", 124),
        B("KEY_LEFTMETA", 125),
        B("KEY_RIGHTMETA", 126),
        B("KEY_COMPOSE", 127),
        B("KEY_STOP", 128),
        B("KEY_AGAIN", 129),
        B("KEY_PROPS", 130),
        B("KEY_UNDO", 131),
        B("KEY_FRONT", 132),
        B("KEY_COPY", 133),
        B("KEY_OPEN", 134),
        B("KEY_PASTE", 135),
        B("KEY_FIND", 136),
        B("KEY_CUT", 137),
        B("KEY_HELP", 138),
        B("KEY_CALC", 140),
        B("KEY_SLEEP", 142),
        B("KEY_WWW", 150),
        B(["KEY_COFFEE", "KEY_SCREENLOCK"], 152),
        B("KEY_BACK", 158),
        B("KEY_FORWARD", 159),
        B("KEY_EJECTCD", 161),
        B("KEY_NEXTSONG", 163),
        B("KEY_PLAYPAUSE", 164),
        B("KEY_PREVIOUSSONG", 165),
        B("KEY_STOPCD", 166),
        B("KEY_REFRESH", 173),
        B("KEY_EDIT", 176),
        B("KEY_SCROLLUP", 177),
        B("KEY_SCROLLDOWN", 178),
        B("KEY_KPLEFTPAREN", 179),
        B("KEY_KPRIGHTPAREN", 180),
        B("KEY_F13", 183),
        B("KEY_F14", 184),
        B("KEY_F15", 185),
        B("KEY_F16", 186),
        B("KEY_F17", 187),
        B("KEY_F18", 188),
        B("KEY_F19", 189),
        B("KEY_F20", 190),
        B("KEY_F21", 191),
        B("KEY_F22", 192),
        B("KEY_F23", 193),
        B("KEY_F24", 194),
        B("KEY_UNKNOWN", 240),
    ],
    B("EV_MSC", 4): [B("MSC_SCAN", 4)],
    B("EV_LED", 17): [B("LED_NUML", 0), B("LED_CAPSL", 1), B("LED_SCROLLL", 2)],
}

MOUSE_CAPABILITIES = {
    # B("EV_SYN", 0): [
    #     B("SYN_REPORT", 0),
    #     B("SYN_CONFIG", 1),
    #     B("SYN_MT_REPORT", 2),
    #     B("?", 4),
    # ],
    B("EV_KEY", 1): [
        B(["BTN_LEFT", "BTN_MOUSE"], 272),
        B("BTN_RIGHT", 273),
        B("BTN_MIDDLE", 274),
        B("BTN_SIDE", 275),
        B("BTN_EXTRA", 276),
    ],
    B("EV_REL", 2): [
        B("REL_X", 0),
        B("REL_Y", 1),
        B("REL_WHEEL", 8),
        B("REL_WHEEL_HI_RES", 11),
    ],
    B("EV_MSC", 4): [B("MSC_SCAN", 4)],
}

TOUCHPAD_CAPABILITIES = {
    B("EV_KEY", 1): [
        B(["BTN_LEFT", "BTN_MOUSE"], 272),
        B("BTN_RIGHT"),
        B("BTN_TOOL_FINGER", 325),
        B("BTN_TOUCH", 330),
        B("BTN_TOOL_DOUBLETAP", 333),
        B("BTN_TOOL_TRIPLETAP", 334),
    ],
    B("EV_ABS", 3): [
        (
            B("ABS_X", 0),
            AbsInfo(value=172, min=0, max=2048, fuzz=0, flat=0, resolution=36),
        ),
        (
            B("ABS_Y", 1),
            AbsInfo(value=472, min=0, max=2048, fuzz=0, flat=0, resolution=36),
        ),
        (
            B("ABS_MT_SLOT", 47),
            AbsInfo(value=0, min=0, max=2, fuzz=0, flat=0, resolution=0),
        ),
        (
            B("ABS_MT_POSITION_X", 53),
            AbsInfo(value=0, min=0, max=2048, fuzz=0, flat=0, resolution=36),
        ),
        (
            B("ABS_MT_POSITION_Y", 54),
            AbsInfo(value=0, min=0, max=2048, fuzz=0, flat=0, resolution=36),
        ),
        (
            B("ABS_MT_TOOL_TYPE", 55),
            AbsInfo(value=0, min=0, max=2, fuzz=0, flat=0, resolution=0),
        ),
        (
            B("ABS_MT_TRACKING_ID", 57),
            AbsInfo(value=0, min=0, max=65535, fuzz=0, flat=0, resolution=0),
        ),
    ],
    B("EV_MSC", 4): [B("MSC_TIMESTAMP", 5)],
}


GAMEPAD_BUTTON_MAP: dict[Button, int] = {
    # Gamepad
    "a": B("BTN_A"),
    "b": B("BTN_B"),
    "x": B("BTN_X"),
    "y": B("BTN_Y"),
    # Sticks
    "ls": B("BTN_THUMBL"),
    "rs": B("BTN_THUMBR"),
    # Bumpers
    "lb": B("BTN_TL"),
    "rb": B("BTN_TR"),
    # Select
    "start": B("BTN_START"),
    "select": B("BTN_SELECT"),
    # Misc
    "mode": B("BTN_MODE"),
    "share": B("BTN_TRIGGER_HAPPY20"),
    # Back buttons
    "extra_l1": B("BTN_TRIGGER_HAPPY1"),
    "extra_r1": B("BTN_TRIGGER_HAPPY2"),
    "extra_l2": B("BTN_TRIGGER_HAPPY3"),
    "extra_r2": B("BTN_TRIGGER_HAPPY4"),
    "extra_l3": B("BTN_TRIGGER_HAPPY5"),
    "extra_r3": B("BTN_TRIGGER_HAPPY6"),
}

GAMEPAD_AXIS_MAP: dict[Axis, AX] = {
    "ls_x": AX(B("ABS_X"), 2**15 - 1),
    "ls_y": AX(B("ABS_Y"), 2**15 - 1),
    "rs_x": AX(B("ABS_RX"), 2**15 - 1),
    "rs_y": AX(B("ABS_RY"), 2**15 - 1),
    "rt": AX(B("ABS_Z"), 2**8 - 1),
    "lt": AX(B("ABS_RZ"), 2**8 - 1),
    "hat_x": AX(B("ABS_HAT0X")),
    "hat_y": AX(B("ABS_HAT0Y")),
}

MOTION_AXIS_MAP: dict[Axis, AX] = {
    "accel_x": AX(B("ABS_X"), 8192, bounds=(-32768, 32768)),
    "accel_y": AX(B("ABS_Y"), 8192, bounds=(-32768, 32768)),
    "accel_z": AX(B("ABS_Z"), 8192, bounds=(-32768, 32768)),
    "gyro_x": AX(B("ABS_RX"), 1024, bounds=(-2097152, 2097152)),
    "gyro_y": AX(B("ABS_RY"), 1024, bounds=(-2097152, 2097152)),
    "gyro_z": AX(B("ABS_RZ"), 1024, bounds=(-2097152, 2097152)),
}

TOUCHPAD_AXIS_MAP: dict[Axis, AX] = {
    "touchpad_x": AX(B("ABS_X"), 1023, bounds=(0, 2048)),
    "touchpad_y": AX(B("ABS_Y"), 1023, bounds=(0, 2048)),
}

TOUCHPAD_BUTTON_MAP: dict[Button, int] = {
    "touchpad_touch": B("BTN_TOUCH"),
    "touchpad_right": B("BTN_RIGHT"),
    "touchpad_left": B("BTN_LEFT"),
}



import logging
import select
import time
from typing import Any, Literal, NamedTuple, Sequence, TypedDict

from .const import Axis, Button, Configuration

logger = logging.getLogger(__name__)


class RumbleEvent(TypedDict):
    """In case ev effects is too complicated. If both magnitudes are 0, disable rumble."""

    type: Literal["rumble"]
    code: Literal["main", "left", "right"]
    strong_magnitude: float
    weak_magnitude: float


class RgbLedEvent(TypedDict):
    """Inspired by new controllers with RGB leds, especially below the buttons.

    Instead of code, this event type exposes multiple properties, including mode."""

    type: Literal["led"]

    # The led
    code: Literal["main", "left", "right"]

    # Various lighting modes supported by the led.
    mode: Literal["disable", "solid", "blinking", "rainbow", "spiral"]

    # Brightness range is from 0 to 1
    # If the response device does not support brightness control, it shall
    # convert the rgb color to hue, assume saturation is 1, and derive a new
    # RGB value from the brightness below
    brightness: float

    # The speed the led should blink if supported by the led
    speed: float

    # Color values for the led, may be ommited depending on the mode, by being
    # set to 0
    red: int
    green: int
    blue: int


class ButtonEvent(TypedDict):
    type: Literal["button"]
    code: Button
    value: bool


class AxisEvent(TypedDict):
    type: Literal["axis"]
    code: Axis
    value: float


class ConfigurationEvent(TypedDict):
    type: Literal["configuration"]
    code: Configuration
    value: Any


Event = ButtonEvent | AxisEvent | ConfigurationEvent | RgbLedEvent | RumbleEvent


class TouchpadCorrection(NamedTuple):
    x_mult: float = 1
    x_ofs: float = 0
    x_clamp: tuple[float, float] = (0, 1)
    y_mult: float = 1
    y_ofs: float = 0
    y_clamp: tuple[float, float] = (0, 1)


TouchpadCorrectionType = Literal[
    "stretch",
    "crop_center",
    "crop_start",
    "crop_end",
    "contain_start",
    "contain_end",
    "contain_center",
    "disabled",
]


def correct_touchpad(
    width: int, height: int, aspect: float, method: TouchpadCorrectionType
):
    dst = width / height
    src = aspect
    ratio = dst / src

    match method:
        case "crop_center":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=(width - new_width) / 2,
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=(height - new_height) / 2,
                )
        case "crop_start":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=0,
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=0,
                )
        case "crop_end":
            if ratio > 1:
                new_width = width / ratio
                return TouchpadCorrection(
                    x_mult=new_width,
                    x_ofs=(width - new_width),
                    y_mult=height,
                    y_ofs=0,
                )
            else:
                new_height = height * ratio
                return TouchpadCorrection(
                    x_mult=width,
                    x_ofs=0,
                    y_mult=new_height,
                    y_ofs=(height - new_height),
                )
        case "contain_center":
            if ratio > 1:
                bound = (ratio - 1) / ratio / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(bound, 1 - bound)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(bound, 1 - bound)
                )
        case "contain_start":
            if ratio > 1:
                bound = (ratio - 1) / ratio
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(0, 1 - bound)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(0, 1 - bound)
                )
        case "contain_end":
            if ratio > 1:
                bound = (ratio - 1) / ratio
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, y_clamp=(bound, 1)
                )
            else:
                bound = (1 - ratio) / 2
                return TouchpadCorrection(
                    x_mult=width, y_mult=height, x_clamp=(bound, 1)
                )
        case "stretch" | "disabled":
            return TouchpadCorrection(x_mult=width, y_mult=height)

    logger.error(f"Touchpad correction method '{method}' not found.")
    return TouchpadCorrection(x_mult=width, y_mult=height)


class Producer:
    def open(self) -> Sequence[int]:
        """Opens and returns a list of file descriptors that should be listened to."""
        raise NotImplementedError()

    def close(self, exit: bool) -> bool:
        """Called to close the device.

        If `exit` is true, the program is about to
        close. If it is false, the controller is entering power save mode because
        it is unused. In this case, if this service is required, you may forgo
        closing and return false. If true, it is assumed this producer is closed.

        `open()` will be called again once the consumers are ready."""
        return False

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        """Called with the file descriptors that are ready to read."""
        return []


class Consumer:
    available: bool
    """Hint that states if the consumer can receive events. If it is false,
    consumer will not be called. If all consumers are false, producers will
    be closed to save CPU utilisation."""

    def initialize(self):
        """Optional method for initialization."""
        pass

    def consume(self, events: Sequence[Event]):
        pass


TouchpadAction = Literal["disabled", "left_click", "right_click"]


class Multiplexer:
    QAM_DELAY = 0.2
    REBOOT_HOLD = 4
    REBOOT_VIBRATION_STRENGTH = 0.6
    REBOOT_VIBRATION_ON = 0.3
    REBOOT_VIBRATION_OFF = 0.8
    REBOOT_VIBRATION_NUM = 3

    def __init__(
        self,
        swap_guide: None | Literal["guide_is_start", "guide_is_select"] = None,
        trigger: None | Literal["analog_to_discrete", "discrete_to_analogue"] = None,
        dpad: None | Literal["analog_to_discrete"] = None,
        led: None | Literal["left_to_main", "right_to_main", "main_to_sides"] = None,
        touchpad: None
        | Literal["left_to_main", "right_to_main", "main_to_sides"] = None,
        status: None | Literal["both_to_main"] = None,
        share_to_qam: bool = False,
        trigger_discrete_lvl: float = 0.99,
        touchpad_short: TouchpadAction = "disabled",
        touchpad_right: TouchpadAction = "left_click",
        touchpad_hold: TouchpadAction = "disabled",
        r3_to_share: bool = False,
        select_reboots: bool = False,
        nintendo_mode: bool = False,
        qam_button: str | None = None,
    ) -> None:
        self.swap_guide = swap_guide
        self.trigger = trigger
        self.dpad = dpad
        self.led = led
        self.touchpad = touchpad
        self.status = status
        self.trigger_discrete_lvl = trigger_discrete_lvl
        self.touchpad_short = touchpad_short
        self.touchpad_hold = touchpad_hold
        self.touchpad_right = touchpad_right
        self.select_reboots = select_reboots
        self.r3_to_share = r3_to_share
        self.nintendo_mode = nintendo_mode

        self.state = {}
        self.touchpad_x = 0
        self.touchpad_y = 0
        self.touchpad_down = None
        self.queue: list[tuple[Event | Literal["reboot"], float]] = []
        self.select_pressed = None
        self.select_is_held = False
        self.qam_button = qam_button
        if share_to_qam:
            self.qam_button = "share"

        assert touchpad is None, "touchpad rewiring not supported yet"

    def process(self, events: Sequence[Event]):
        out: list[Event] = []
        status_events = set()
        touched = False

        curr = time.perf_counter()
        while len(self.queue) and self.queue[0][1] < curr:
            ev = self.queue.pop(0)[0]
            if ev == "reboot":
                if self.select_is_held:
                    try:
                        import os

                        os.system("systemctl reboot")
                        logger.info("rebooting")
                    except Exception as e:
                        logger.error(f"Rebooting failed with error:\n{type(e)}:{e}")
            elif self.select_is_held or not ev.get("from_reboot", False):
                out.append(ev)

        if self.select_pressed and self.select_pressed + self.REBOOT_HOLD < curr:
            self.select_pressed = None
            for i in range(self.REBOOT_VIBRATION_NUM):
                self.queue.append(
                    (
                        {
                            "type": "rumble",
                            "code": "main",
                            "strong_magnitude": self.REBOOT_VIBRATION_STRENGTH,
                            "weak_magnitude": self.REBOOT_VIBRATION_STRENGTH,
                            "from_reboot": True,
                        },  # type: ignore
                        curr
                        + i * (self.REBOOT_VIBRATION_ON + self.REBOOT_VIBRATION_OFF),
                    )
                )
                self.queue.append(
                    (
                        {
                            "type": "rumble",
                            "code": "main",
                            "strong_magnitude": 0,
                            "weak_magnitude": 0,
                        },
                        curr
                        + i * (self.REBOOT_VIBRATION_ON + self.REBOOT_VIBRATION_OFF)
                        + self.REBOOT_VIBRATION_ON,
                    )
                )
            self.queue.append(("reboot", curr))

        if (
            self.touchpad_hold != "disabled"
            and self.touchpad_down
            and curr - self.touchpad_down[0] > 0.8
            and abs(self.touchpad_down[1] - self.touchpad_x) < 0.1
            and abs(self.touchpad_down[2] - self.touchpad_y) < 0.1
        ):
            action = (
                "touchpad_left"
                if self.touchpad_hold == "left_click"
                else "touchpad_right"
            )
            self.queue.append(
                (
                    {
                        "type": "button",
                        "code": action,
                        "value": True,
                    },
                    curr,
                )
            )
            self.queue.append(
                (
                    {
                        "type": "button",
                        "code": action,
                        "value": False,
                    },
                    curr + self.QAM_DELAY,
                )
            )
            self.touchpad_down = None

        for ev in events:
            match ev["type"]:
                case "axis":
                    if self.trigger == "analog_to_discrete" and ev["code"] in (
                        "lt",
                        "rt",
                    ):
                        out.append(
                            {
                                "type": "button",
                                "code": ev["code"],
                                "value": ev["value"] > self.trigger_discrete_lvl,
                            }
                        )

                    if self.dpad == "analog_to_discrete" and ev["code"] in (
                        "hat_x",
                        "hat_y",
                    ):
                        out.append(
                            {
                                "type": "button",
                                "code": "dpad_up"
                                if ev["code"] == "hat_y"
                                else "dpad_right",
                                "value": ev["value"] > 0.5,
                            }
                        )
                        out.append(
                            {
                                "type": "button",
                                "code": "dpad_down"
                                if ev["code"] == "hat_y"
                                else "dpad_left",
                                "value": ev["value"] < -0.5,
                            }
                        )
                    if ev["code"] == "touchpad_x":
                        self.touchpad_x = ev["value"]
                    if ev["code"] == "touchpad_y":
                        self.touchpad_y = ev["value"]
                case "button":
                    if self.trigger == "discrete_to_analog" and ev["code"] in (
                        "lt",
                        "rt",
                    ):
                        out.append(
                            {
                                "type": "axis",
                                "code": ev["code"],
                                "value": 1 if ev["value"] else 0,
                            }
                        )

                    if self.swap_guide and ev["code"] in (
                        "start",
                        "select",
                        "mode",
                        "share",
                    ):
                        match ev["code"]:
                            case "start":
                                ev["code"] = "mode"
                            case "select":
                                ev["code"] = "share"
                            case "mode":
                                if self.swap_guide == "guide_is_start":
                                    ev["code"] = "start"
                                else:
                                    ev["code"] = "select"
                            case "share":
                                if self.swap_guide == "guide_is_start":
                                    ev["code"] = "select"
                                else:
                                    ev["code"] = "start"

                    if self.select_reboots and ev["code"] == "select":
                        if ev["value"]:
                            self.select_is_held = True
                            self.select_pressed = curr
                        else:
                            self.select_is_held = False
                            self.select_pressed = None

                    if self.qam_button is not None and ev["code"] == self.qam_button:
                        if ev["value"]:
                            ev["code"] = "mode"
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "a",
                                        "value": True,
                                    },
                                    curr + self.QAM_DELAY,
                                )
                            )
                        else:
                            # TODO: Clean this up
                            ev["code"] = ""  # type: ignore
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "mode",
                                        "value": False,
                                    },
                                    curr + self.QAM_DELAY,
                                ),
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": "a",
                                        "value": False,
                                    },
                                    curr + self.QAM_DELAY,
                                ),
                            )

                    if ev["code"] == "touchpad_right":
                        match self.touchpad_right:
                            case "disabled":
                                # TODO: Cleanup
                                ev["code"] = ""  # type: ignore
                            case "left_click":
                                ev["code"] = "touchpad_left"
                            case "right_click":
                                pass

                    if ev["code"] == "touchpad_touch":
                        if (
                            self.touchpad_short != "disabled"
                            and not ev["value"]
                            and self.touchpad_down
                            and curr - self.touchpad_down[0] < 0.2
                            and abs(self.touchpad_down[1] - self.touchpad_x) < 0.04
                            and abs(self.touchpad_down[2] - self.touchpad_y) < 0.04
                        ):
                            action = (
                                "touchpad_left"
                                if self.touchpad_short == "left_click"
                                else "touchpad_right"
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": action,
                                        "value": True,
                                    },
                                    curr,
                                )
                            )
                            self.queue.append(
                                (
                                    {
                                        "type": "button",
                                        "code": action,
                                        "value": False,
                                    },
                                    curr + self.QAM_DELAY,
                                )
                            )

                        if ev["value"]:
                            touched = True
                        else:
                            self.touchpad_down = None
                        # append A after QAM_DELAY s

                    if self.r3_to_share and ev["code"] == "extra_r3":
                        ev["code"] = "share"

                    if self.nintendo_mode:
                        match ev["code"]:
                            case "a":
                                ev["code"] = "b"
                            case "b":
                                ev["code"] = "a"
                            case "x":
                                ev["code"] = "y"
                            case "y":
                                ev["code"] = "x"
                case "led":
                    if self.led == "left_to_main" and ev["code"] == "left":
                        out.append({**ev, "code": "main"})
                    elif self.led == "right_to_main" and ev["code"] == "right":
                        out.append({**ev, "code": "main"})
                    elif self.led == "main_to_both" and ev["code"] == "main":
                        out.append({**ev, "code": "left"})
                        out.append({**ev, "code": "right"})
                case "configuration":
                    if self.status == "both_to_main":
                        self.state[ev["code"]] = ev["value"]
                        match ev["code"]:
                            case "battery_left" | "battery_right":
                                status_events.add("battery")
                            case "is_attached_left" | "is_attached_right":
                                status_events.add("is_attached")
                            case "is_connected_left" | "is_connected_right":
                                status_events.add("is_connected")

        if touched:
            self.touchpad_down = (
                curr,
                self.touchpad_x,
                self.touchpad_y,
            )

        for s in status_events:
            match s:
                case "battery":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "battery",
                            "value": min(
                                self.state.get("battery_left", 100),
                                self.state.get("battery_right", 100),
                            ),
                        }
                    )
                case "is_attached":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "is_attached",
                            "value": (
                                self.state.get("is_attached_left", False)
                                and self.state.get("is_attached_right", False)
                            ),
                        }
                    )
                case "is_connected":
                    out.append(
                        {
                            "type": "configuration",
                            "code": "is_connected",
                            "value": (
                                self.state.get("is_connected_left", False)
                                and self.state.get("is_connected_right", False)
                            ),
                        }
                    )

        out.extend(events)
        return out


class KeyboardWrapper(Producer, Consumer):
    def __init__(
        self, parent: Producer, button_map: Sequence[tuple[set[Button], Button]]
    ) -> None:
        self.parent = parent
        self.button_map = button_map

        self.active_in: set[Button] = set()
        self.active_out: set[Button] = set()

    def open(self) -> Sequence[int]:
        self.active_in: set[Button] = set()
        self.active_out: set[Button] = set()
        return self.parent.open()

    def close(self, exit: bool) -> bool:
        return self.parent.close(exit)

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        evs: Sequence[Event] = self.parent.produce(fds)
        # Update in map
        for ev in evs:
            logger.info(f"Internal kbd event: {ev}")
            if ev["type"] == "button":
                if ev["value"]:
                    self.active_in.add(ev["code"])
                elif ev["code"] in self.active_in:
                    self.active_in.remove(ev["code"])

        # Debounce and output
        out: Sequence[Event] = []
        for bset, action in self.button_map:
            is_sub = bset.issubset(self.active_in)
            is_active = action in self.active_out
            if is_sub and not is_active:
                self.active_out.add(action)
                out.append({"type": "button", "code": action, "value": True})
            elif not is_sub and is_active:
                self.active_out.remove(action)
                out.append({"type": "button", "code": action, "value": False})

        return out

    def consume(self, events: Sequence[Event]):
        if isinstance(self.parent, Consumer):
            return self.parent.consume(events)


def can_read(fd: int):
    return select.select([fd], [], [], 0)[0]



import logging
import time
from collections import defaultdict
from typing import Sequence

from hhd.controller import Consumer, Event, Producer
from hhd.controller.lib.uhid import UhidDevice, BUS_USB

from .const import (
    SDCONT_VENDOR,
    SDCONT_PRODUCT,
    SDCONT_VERSION,
    SDCONT_COUNTRY,
    SDCONT_NAME,
    SDCONT_DESCRIPTOR,
)

logger = logging.getLogger(__name__)


class SteamdeckOLEDController(Producer, Consumer):
    def __init__(
        self,
    ) -> None:
        self.available = False
        self.report = None
        self.dev = None
        self.start = 0

    def open(self) -> Sequence[int]:
        self.available = False
        self.report = bytearray([64] + [0 for _ in range(64)])
        self.dev = UhidDevice(
            vid=SDCONT_VENDOR,
            pid=SDCONT_PRODUCT,
            bus=BUS_USB,
            version=SDCONT_VERSION,
            country=SDCONT_COUNTRY,
            name=SDCONT_NAME,
            report_descriptor=SDCONT_DESCRIPTOR,
        )

        self.state: dict = defaultdict(lambda: 0)
        self.rumble = False
        self.touchpad_touch = False
        self.start = time.perf_counter_ns()
        self.fd = self.dev.open()
        return [self.fd]

    def close(self, exit: bool) -> bool:
        if not exit:
            """This is a consumer, so we would deadlock if it was disabled."""
            return False

        if self.dev:
            self.dev.send_destroy()
            self.dev.close()
            self.dev = None
            self.fd = None

        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        return []

    def consume(self, events: Sequence[Event]):
        pass



import logging
import select
import time
from threading import Event as TEvent
from typing import Literal, Sequence

from hhd.controller import Axis, Event, Multiplexer, can_read
from hhd.controller.base import Event
from hhd.controller.lib.common import AM, BM, CM
from hhd.controller.lib.hid import MAX_REPORT_SIZE
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev
from hhd.controller.physical.hidraw import EventCallback, GenericGamepadHidraw
from hhd.controller.physical.imu import CombinedImu, HrtimerTrigger
from hhd.plugins import Config, Context, Emitter, get_outputs

from .hid import RgbCallback, switch_mode, Brightness

ERROR_DELAY = 1
SELECT_TIMEOUT = 1

logger = logging.getLogger(__name__)

ASUS_VID = 0x0B05
ASUS_KBD_PID = 0x1ABE
GAMEPAD_VID = 0x045E
GAMEPAD_PID = 0x028E

ALLY_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("gyro_x", "accel", 1, 3),
    "accel_y": ("gyro_z", "accel", -1, 3),
    "accel_z": ("gyro_y", "accel", -1, 3),
    "anglvel_x": ("gyro_x", "anglvel", 1, None),
    "anglvel_y": ("gyro_z", "anglvel", -1, None),
    "anglvel_z": ("gyro_y", "anglvel", -1, None),
    "timestamp": ("gyro_ts", None, 1, None),
}

MODE_DELAY = 0.3
VIBRATION_DELAY = 0.1
VIBRATION_ON: Event = {
    "type": "rumble",
    "code": "main",
    "strong_magnitude": 0.5,
    "weak_magnitude": 0.5,
}
VIBRATION_OFF: Event = {
    "type": "rumble",
    "code": "main",
    "strong_magnitude": 0,
    "weak_magnitude": 0,
}


class AllyHidraw(GenericGamepadHidraw):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def open(self) -> Sequence[int]:
        self.queue: list[tuple[Event, float]] = []
        a = super().open()
        if self.dev:
            logger.info(f"Switching Ally Controllers to gamepad mode.")
            switch_mode(self.dev, "default")

        self.mouse_mode = False
        return a

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        # If we can not read return
        if not self.fd or not self.dev:
            return []

        # Process events
        curr = time.perf_counter()
        out: Sequence[Event] = []

        # Remove up to one queued event
        if len(self.queue):
            ev, ofs = self.queue[0]
            if ofs < curr:
                out.append(ev)
                self.queue.pop(0)

        # Read new events
        while can_read(self.fd):
            rep = self.dev.read(self.report_size)
            # logger.warning(f"Received the following report (debug):\n{rep.hex()}")
            if rep[0] != 0x5A:
                continue

            match rep[1]:
                case 0xA6:
                    # action = "left"
                    out.append({"type": "button", "code": "mode", "value": True})
                    self.queue.append(
                        (
                            {"type": "button", "code": "mode", "value": False},
                            curr + MODE_DELAY,
                        )
                    )
                case 0x38:
                    # action = "right"
                    out.append({"type": "button", "code": "share", "value": True})
                    self.queue.append(
                        (
                            {"type": "button", "code": "share", "value": False},
                            curr + MODE_DELAY,
                        )
                    )
                case 0xA7:
                    # right hold
                    # Mode switch
                    if self.mouse_mode:
                        switch_mode(self.dev, "default")
                        self.mouse_mode = False
                        out.append(VIBRATION_ON)
                        self.queue.append((VIBRATION_OFF, curr + VIBRATION_DELAY))
                    else:
                        switch_mode(self.dev, "mouse")
                        self.mouse_mode = True
                        out.append(VIBRATION_ON)
                        self.queue.append((VIBRATION_OFF, curr + VIBRATION_DELAY))
                        self.queue.append((VIBRATION_ON, curr + 2 * VIBRATION_DELAY))
                        self.queue.append((VIBRATION_OFF, curr + 3 * VIBRATION_DELAY))
                case 0xA8:
                    # action = "right_hold_release"
                    pass  # kind of useless

        return out


def plugin_run(
    conf: Config, emit: Emitter, context: Context, should_exit: TEvent, updated: TEvent
):
    while not should_exit.is_set():
        try:
            logger.info("Launching emulated controller.")
            updated.clear()
            controller_loop(conf.copy(), should_exit, updated)
        except Exception as e:
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {ERROR_DELAY}s."
            )
            # Raise exception
            if conf.get("debug", False):
                raise e
            time.sleep(ERROR_DELAY)


def controller_loop(conf: Config, should_exit: TEvent, updated: TEvent):
    debug = conf.get("debug", False)

    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"], None, conf["imu"].to(bool)
    )

    # Imu
    d_imu = CombinedImu(conf["imu_hz"].to(int), ALLY_MAPPINGS, gyro_scale="0.000266")
    d_timer = HrtimerTrigger(conf["imu_hz"].to(int), [HrtimerTrigger.IMU_NAMES])

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[GAMEPAD_VID],
        pid=[GAMEPAD_PID],
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
    )

    # Vendor
    d_vend = AllyHidraw(
        vid=[ASUS_VID],
        pid=[ASUS_KBD_PID],
        usage_page=[0xFF31],
        usage=[0x0080],
        required=True,
        callback=RgbCallback(conf["led_brightness"].to(Brightness)),
    )

    # Grab shortcut keyboards
    d_kbd_1 = GenericGamepadEvdev(
        vid=[ASUS_VID],
        pid=[ASUS_KBD_PID],
        capabilities={EC("EV_KEY"): [EC("KEY_F23")]},
        required=False,
        grab=False,
        btn_map={EC("KEY_F17"): "extra_l1", EC("KEY_F18"): "extra_r1"},
    )

    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        share_to_qam=conf["share_to_qam"].to(bool),
        select_reboots=conf["select_reboots"].to(bool),
        nintendo_mode=conf["nintendo_mode"].to(bool),
    )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    if conf["imu"].to(bool):
        REPORT_FREQ_MAX = max(REPORT_FREQ_MAX, conf["imu_hz"].to(float))

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        d_vend.open()
        prepare(d_xinput)
        if conf.get("imu", False):
            if d_timer.open():
                prepare(d_imu)
        prepare(d_kbd_1)
        for d in d_producers:
            prepare(d)

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))
            evs.extend(d_vend.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

            d_vend.consume(evs)
            d_xinput.consume(evs)

            for d in d_outs:
                d.consume(evs)

            # If unbounded, the total number of events per second is the sum of all
            # events generated by the producers.
            # For Legion go, that would be 100 + 100 + 500 + 30 = 730
            # Since the controllers of the legion go only update at 500hz, this is
            # wasteful.
            # By setting a target refresh rate for the report and sleeping at the
            # end, we ensure that even if multiple fds become ready close to each other
            # they are combined to the same report, limiting resource use.
            # Ideally, this rate is smaller than the report rate of the hardware controller
            # to ensure there is always a report from that ready during refresh
            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        d_vend.close(True)
        d_timer.close()
        for d in reversed(devs):
            d.close(True)



from threading import Event, Thread
from typing import Any, Sequence

from hhd.controller.physical.rgb import is_led_supported
from hhd.plugins import (
    Config,
    Context,
    Emitter,
    HHDPlugin,
    get_gyro_config,
    get_outputs_config,
    load_relative_yaml,
)
from hhd.plugins.settings import HHDSettings

from .const import CONFS, DEFAULT_MAPPINGS, get_default_config


class GenericControllersPlugin(HHDPlugin):
    name = "generic_controllers"
    priority = 18
    log = "genc"

    def __init__(self, dmi: str, dconf: dict) -> None:
        self.t = None
        self.should_exit = None
        self.updated = Event()
        self.started = False
        self.t = None

        self.dmi = dmi
        self.dconf = dconf
        self.name = f"generic_controllers@'{dconf.get('name', 'ukn')}'"

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        self.emit = emit
        self.context = context
        self.prev = None

    def settings(self) -> HHDSettings:
        base = {"controllers": {"handheld": load_relative_yaml("controllers.yml")}}
        base["controllers"]["handheld"]["children"]["controller_mode"].update(
            get_outputs_config(
                can_disable=True,
                has_leds=is_led_supported(),
                start_disabled=self.dconf.get("untested", False),
            )
        )

        base["controllers"]["handheld"]["children"]["imu_axis"] = get_gyro_config(
            self.dconf.get("mapping", DEFAULT_MAPPINGS)
        )

        return base

    def update(self, conf: Config):
        new_conf = conf["controllers.handheld"]
        if new_conf == self.prev:
            return
        if self.prev is None:
            self.prev = new_conf
        else:
            self.prev.update(new_conf.conf)

        self.updated.set()
        self.start(self.prev)

    def start(self, conf):
        from .base import plugin_run

        if self.started:
            return
        self.started = True

        self.close()
        self.should_exit = Event()
        self.t = Thread(
            target=plugin_run,
            args=(
                conf,
                self.emit,
                self.context,
                self.should_exit,
                self.updated,
                self.dconf,
            ),
        )
        self.t.start()

    def close(self):
        if not self.should_exit or not self.t:
            return
        self.should_exit.set()
        self.t.join()
        self.should_exit = None
        self.t = None


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    # Match just product name
    # if a device exists here its officially supported
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        dmi = f.read().strip()

    dconf = CONFS.get(dmi, None)
    if dconf:
        return [GenericControllersPlugin(dmi, dconf)]

    # Begin hw agnostic dmi match
    if "ONEXPLAYER" in dmi:
        return [GenericControllersPlugin(dmi, get_default_config(dmi, "ONEXPLAYER"))]

    try:
        with open("/sys/devices/virtual/dmi/id/sys_vendor") as f:
            vendor = f.read().strip().lower()
        if vendor == "ayn":
            return [GenericControllersPlugin(dmi, get_default_config(dmi, "AYN"))]
    except Exception:
        pass

    # Fallback to chassis vendor for aya
    try:
        with open("/sys/class/dmi/id/board_vendor") as f:
            vendor = f.read().lower().strip()

        if "ayaneo" in vendor:
            return [GenericControllersPlugin(dmi, get_default_config(dmi, "AYA"))]
    except Exception:
        return []

    return []



import logging
import os
import select
import time
from threading import Event as TEvent

import evdev

from hhd.controller import Multiplexer
from hhd.controller.physical.evdev import DINPUT_AXIS_MAP, DINPUT_AXIS_POSTPROCESS
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev
from hhd.controller.physical.imu import CombinedImu, HrtimerTrigger
from hhd.controller.physical.rgb import LedDevice
from hhd.controller.virtual.uinput import UInputDevice
from hhd.plugins import Config, Context, Emitter, get_gyro_state, get_outputs

from .const import BTN_MAPPINGS, DEFAULT_MAPPINGS

ERROR_DELAY = 1
SELECT_TIMEOUT = 1

logger = logging.getLogger(__name__)

GAMEPAD_VIDS = [0x045E, 0x0079]
GAMEPAD_PIDS = [0x028E, 0x181C]

KBD_VID = 0x0001
KBD_PID = 0x0001

BACK_BUTTON_DELAY = 0.1


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
):
    first = True
    while not should_exit.is_set():
        if conf["controller_mode.mode"].to(str) == "disabled":
            time.sleep(ERROR_DELAY)
            continue

        found_gamepad = False
        try:
            for d in evdev.list_devices():
                dev = evdev.InputDevice(d)
                if dev.info.vendor in GAMEPAD_VIDS and dev.info.product in GAMEPAD_PIDS:
                    found_gamepad = True
                    break
        except Exception:
            logger.warning("Failed finding device, skipping check.")
            found_gamepad = True

        if not found_gamepad:
            if first:
                logger.info("Controller not found. Waiting...")
            time.sleep(ERROR_DELAY)
            first = False
            continue

        try:
            logger.info("Launching emulated controller.")
            updated.clear()
            controller_loop(conf.copy(), should_exit, updated, dconf)
        except Exception as e:
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {ERROR_DELAY}s."
            )
            first = True
            # Raise exception
            if conf.get("debug", False):
                raise e
            time.sleep(ERROR_DELAY)


def controller_loop(conf: Config, should_exit: TEvent, updated: TEvent, dconf: dict):
    debug = conf.get("debug", False)

    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        None,
        conf["imu"].to(bool),
    )

    # Imu
    d_imu = CombinedImu(
        conf["imu_hz"].to(int),
        get_gyro_state(conf["imu_axis"], dconf.get("mapping", DEFAULT_MAPPINGS)),
    )
    d_timer = HrtimerTrigger(conf["imu_hz"].to(int), [HrtimerTrigger.IMU_NAMES])

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=GAMEPAD_VIDS,
        pid=GAMEPAD_PIDS,
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        axis_map=DINPUT_AXIS_MAP,
        hide=True,
        postprocess=DINPUT_AXIS_POSTPROCESS,
    )

    d_kbd_1 = GenericGamepadEvdev(
        vid=[KBD_VID],
        pid=[KBD_PID],
        required=False,
        # grab=True,
        btn_map=dconf.get("btn_mapping", BTN_MAPPINGS),
    )

    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        share_to_qam=conf["share_to_qam"].to(bool),
        nintendo_mode=conf["nintendo_mode"].to(bool),
    )

    # d_volume_btn = UInputDevice(
    #     name="Handheld Daemon Volume Keyboard",
    #     phys="phys-hhd-vbtn",
    #     capabilities={EC("EV_KEY"): [EC("KEY_VOLUMEUP"), EC("KEY_VOLUMEDOWN")]},
    #     btn_map={
    #         "key_volumeup": EC("KEY_VOLUMEUP"),
    #         "key_volumedown": EC("KEY_VOLUMEDOWN"),
    #     },
    #     pid=KBD_PID,
    #     vid=KBD_VID,
    #     output_timestamps=True,
    # )

    d_rgb = LedDevice()
    if d_rgb.supported:
        logger.info(f"RGB Support activated through kernel driver.")

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    if conf["imu"].to(bool):
        REPORT_FREQ_MAX = max(REPORT_FREQ_MAX, conf["imu_hz"].to(float))

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        # d_vend.open()
        prepare(d_xinput)
        if conf.get("imu", False):
            start_imu = True
            if dconf.get("hrtimer", False):
                start_imu = d_timer.open()
            if start_imu:
                prepare(d_imu)
        # prepare(d_volume_btn)
        prepare(d_kbd_1)
        for d in d_producers:
            prepare(d)

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                # d_volume_btn.consume(evs)
                d_xinput.consume(evs)

            d_rgb.consume(evs)
            for d in d_outs:
                d.consume(evs)

            # If unbounded, the total number of events per second is the sum of all
            # events generated by the producers.
            # For Legion go, that would be 100 + 100 + 500 + 30 = 730
            # Since the controllers of the legion go only update at 500hz, this is
            # wasteful.
            # By setting a target refresh rate for the report and sleeping at the
            # end, we ensure that even if multiple fds become ready close to each other
            # they are combined to the same report, limiting resource use.
            # Ideally, this rate is smaller than the report rate of the hardware controller
            # to ensure there is always a report from that ready during refresh
            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        # d_vend.close(True)
        d_timer.close()
        for d in reversed(devs):
            d.close(True)
